# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures for testing the SQLAlchemy backend."""

import datetime

import pytest

from nark.tests.backends.sqlalchemy.conftest import *  # noqa: F401, F403


@pytest.fixture(
    params=[
        # 2018-05-05: (lb): I'm so confused. Why is datetime.datetime returned
        # in one test, but freezegun.api.FakeDatetime returned in another?
        # And then for today's date, you use datetime.time! So strange!!
        #
        # I thought it might be because @freezegun.freeze_time only freezes
        # now(), so all other times are not mocked, but note how the frozen
        # 18:00 time, '2015-12-12 18:00', is not mocked in the first two tests,
        # but then it is in the third test!! So confused!
        (
            None,
            None,
            "",
            {
                "since": None,
                "until": None,
            },
        ),
        (
            "2015-12-12 18:00",
            "2015-12-12 19:30",
            "",
            {
                "since": datetime.datetime(2015, 12, 12, 18, 0, 0),
                "until": datetime.datetime(2015, 12, 12, 19, 30, 0),
            },
        ),
        (
            "2015-12-12 18:00",
            "2015-12-12 19:30",
            "",
            {
                "since": datetime.datetime(2015, 12, 12, 18, 0, 0),
                "until": datetime.datetime(2015, 12, 12, 19, 30, 0),
            },
        ),
        (
            "2015-12-12 18:00",
            "",
            "",
            {
                # (lb): Note sure diff btw. FakeDatetime and datetime.
                # 'since': freezegun.api.FakeDatetime(2015, 12, 12, 18, 0, 0),
                "since": datetime.datetime(2015, 12, 12, 18, 0, 0),
                "until": None,  # `not until` becomes None, so '' => None.
            },
        ),
        (
            "2015-12-12",
            "",
            "",
            {
                # 'since': freezegun.api.FakeDatetime(2015, 12, 12, 0, 0, 0),
                "since": datetime.datetime(2015, 12, 12, 0, 0, 0),
                "until": None,  # `not until` becomes None, so '' => None.
            },
        ),
        (
            "13:00",
            "",
            "",
            {
                "since": datetime.datetime(2015, 12, 12, 13, 0, 0),
                "until": None,  # `not until` becomes None, so '' => None.
            },
        ),
    ]
)
def search_parameter_parametrized(request):
    """Provide a parametrized set of arguments for the ``get_all`` command."""
    return request.param



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import json


class TestJSONWriter(object):
    """Make sure the JSON writer works as expected."""

    def test_json_writer_write_facts(self, json_writer, list_of_facts):
        """Make sure the calendar is actually written do disk before file is closed."""
        number_of_facts = 5
        facts = list_of_facts(number_of_facts)
        output_path = json_writer.output_file.name
        json_writer.write_facts(facts)
        with open(output_path, "r") as fobj:
            result = json.load(fobj)
            for idx, fact in enumerate(facts):
                assert result[idx] == json_writer.fact_as_dict(fact)

    def test_json_writer_write_report(self, json_writer, fact, table, headers):
        """Make sure the calendar is actually written do disk before file is closed."""
        output_path = json_writer.output_file.name
        json_writer.write_report(table, headers)
        with open(output_path, "r") as fobj:
            result = json.load(fobj)
            for idx, row in enumerate(table):
                kvals = {key: value for key, value in zip(headers, row)}
                assert result[idx] == kvals



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Base class for CSVWriter and TSVWriter output formats."""

import csv
from gettext import gettext as _

from . import ReportWriter

__all__ = ("PlaintextWriter",)


class PlaintextWriter(ReportWriter):
    def __init__(
        self,
        output_b=False,
        # dialect and **fmtparams passed to csv.writer().
        dialect="excel",
        **fmtparams
    ):
        """
        Initialize a new instance.

        Besides our default behaviour we create a localized heading.
        Also, we need to make sure that our heading is UTF-8 encoded on python 2!
        In that case ``self.output_file`` will be opened in binary mode and will
        accept those encoded headings.
        """
        super(PlaintextWriter, self).__init__(output_b=output_b)
        self.dialect = dialect
        self.fmtparams = fmtparams

    def output_setup(self, *args, **kwargs):
        super(PlaintextWriter, self).output_setup(*args, **kwargs)
        # Note that csv only requires that csvfile has a write() method.
        # Note that dialects are loaded at runtime and the list is not
        # documented other than to show the default dialect is 'excel'.
        #   >>> csv.list_dialects()
        #   ['excel', 'excel-tab', 'unix']
        self.csv_writer = csv.writer(
            self.output_file,
            dialect=self.dialect,
            **self.fmtparams,
        )

    def open_file(self, path, output_b=False, newline=""):
        # Per docs: "If csvfile is a file object, it should be opened with newline=''".
        return super(PlaintextWriter, self).open_file(
            path=path,
            output_b=output_b,
            newline=newline,
        )

    # ***

    def write_facts(self, facts):
        self.csv_writer.writerow(self.facts_headers())
        return super(PlaintextWriter, self).write_facts(facts)

    def _write_fact(self, idx, fact):
        """
        Write a single fact.

        On python 2 we need to make sure we encode our data accordingly so we
        can feed it to our file object which in this case needs to be opened in
        binary mode.
        """
        results = []
        for value in self.fact_as_tuple(fact):
            results.append(value)
        self.csv_writer.writerow(results)

    def facts_headers(self):
        """Export a tuple indicating the report column headers.

        Note that _report_headers and _report_row return matching
        sequences of Fact attributes.
        """
        headers = (
            _("Start time"),
            _("End time"),
            _("Duration"),
            _("Activity"),
            _("Category"),
            _("Description"),
            _("Deleted"),
        )
        return headers

    def fact_as_tuple(self, fact):
        """Export a tuple indicating a single report row values.

        Note that _report_headers and _report_row return matching
        sequences of Fact attributes.
        """
        row = (
            fact.start_fmt(self.datetime_format),
            fact.end_fmt(self.datetime_format),
            fact.format_delta(style=self.duration_fmt),
            fact.activity_name,
            fact.category_name,
            fact.description_or_empty,
            str(fact.deleted),
        )
        return row

    # ***

    def write_report(self, table, headers, tabulation=None):
        self.csv_writer.writerow(headers)
        return super(PlaintextWriter, self).write_report(table, headers, tabulation)

    def _write_result(self, row, headers, tabulation=None):
        """
        Write a single fact.

        On python 2 we need to make sure we encode our data accordingly so we
        can feed it to our file object which in this case needs to be opened in
        binary mode.
        """
        self.csv_writer.writerow(row)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import pytest


class TestTagManager:
    def test_save_new(self, basestore, tag, mocker):
        """Make sure that saving an new tag calls ``__add``."""
        mocker.patch.object(basestore.tags, "_add", return_value=tag)
        try:
            basestore.tags.save(tag)
        except NotImplementedError:
            pass
        assert basestore.tags._add.called

    def test_save_existing(self, basestore, tag, mocker):
        tag.pk = 0
        mocker.patch.object(basestore.tags, "_update", return_value=tag)
        try:
            basestore.tags.save(tag)
        except NotImplementedError:
            pass
        assert basestore.tags._update.called

    def test_save_wrong_type(self, basestore, tag):
        with pytest.raises(TypeError):
            basestore.tags.save([])

    def test_get_or_create_existing(self, basestore, tag, mocker):
        """Make sure the tag is beeing looked up and no new one is created."""
        mocker.patch.object(basestore.tags, "get_by_name", return_value=tag)
        mocker.patch.object(basestore.tags, "_add", return_value=tag)
        try:
            basestore.tags.get_or_create(tag.name)
        except NotImplementedError:
            pass
        assert basestore.tags._add.called is False
        assert basestore.tags.get_by_name.called

    def test_get_or_create_new_tag(self, basestore, tag, mocker):
        """Make sure the tag is beeing looked up and new one is created."""
        mocker.patch.object(basestore.tags, "_add", return_value=tag)
        mocker.patch.object(basestore.tags, "get_by_name", side_effect=KeyError)
        try:
            basestore.tags.get_or_create(tag.name)
        except NotImplementedError:
            pass
        assert basestore.tags.get_by_name.called
        assert basestore.tags._add.called

    def test_add_not_implemented(self, basestore, tag):
        with pytest.raises(NotImplementedError):
            basestore.tags._add(tag)

    def test_update_not_implemented(self, basestore, tag):
        with pytest.raises(NotImplementedError):
            basestore.tags._update(tag)

    def test_remove_not_implemented(self, basestore, tag):
        with pytest.raises(NotImplementedError):
            basestore.tags.remove(tag)

    def test_get_not_implemented_invalid_pk(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get(12)

    def test_get_not_implemented_invalid_pk_type(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get_by_name("fooo")

    def test_get_all_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get_all()

    def test_get_all_by_usage_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.get_all_by_usage()

    def test_gather_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.tags.gather(query_terms=None)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""nark provides generic time tracking functionality."""

import time

from easy_as_pypi_getver import get_version as _get_version

__all__ = (
    "get_version",
    "__package_name__",
    "__time_0__",
    "__PROFILING__",
    # Private:
    #  '_version_from_tags',
)

__PROFILING__ = True
# DEVS: Comment this out to see load times summary.
__PROFILING__ = False
__time_0__ = time.time()

# (lb): Seems a little redundant (see setup.cfg:[metadata]name)
# but not sure if way to get programmatically. This is closest
# solution that avoids hardcoding the library name in strings
# (which is something linter or runtime won't catch if wrong).
__package_name__ = "nark"

# This version is substituted on poetry-build by poetry-dynamic-versioning.
# - Consequently, __version__ remains empty when installed in 'editable' mode.
__version__ = "3.2.8"


def get_version(include_head=False):
    return _get_version(
        package_name=__package_name__,
        reference_file=__file__,
        include_head=include_head,
    )



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest

from nark.reports import ReportWriter
from nark.reports.csv_writer import CSVWriter
from nark.reports.ical_writer import ICALWriter
from nark.reports.json_writer import JSONWriter
from nark.reports.plaintext_writer import PlaintextWriter
from nark.reports.tsv_writer import TSVWriter
from nark.reports.xml_writer import XMLWriter

# Register the fact_factory, etc.
from nark.tests.item_factories import *  # noqa: F401, F403


@pytest.fixture
def path(tmpdir):
    path = tmpdir.mkdir("reports").join("export.fmt").strpath
    return path


@pytest.fixture
def report_writer(path):
    report_writer = ReportWriter()
    report_writer.output_setup(path)
    return report_writer


@pytest.fixture
def csv_writer(path):
    csv_writer = CSVWriter()
    csv_writer.output_setup(path)
    return csv_writer


@pytest.fixture
def ical_writer(path):
    ical_writer = ICALWriter()
    ical_writer.output_setup(path)
    return ical_writer


@pytest.fixture
def json_writer(path):
    json_writer = JSONWriter()
    json_writer.output_setup(path)
    return json_writer


@pytest.fixture
def plaintext_writer(path):
    plaintext_writer = PlaintextWriter()
    plaintext_writer.output_setup(path)
    return plaintext_writer


@pytest.fixture
def tsv_writer(path):
    tsv_writer = TSVWriter()
    tsv_writer.output_setup(path)
    return tsv_writer


@pytest.fixture
def xml_writer(path):
    xml_writer = XMLWriter()
    xml_writer.output_setup(path)
    return xml_writer


# ***


@pytest.fixture
def list_of_facts(fact_factory):
    """
    Provide a factory that returns a list with given amount of Fact instances.

    The key point here is that these fact *do not overlap*!
    """

    def get_list_of_facts(number_of_facts):
        facts = []
        # MAYBE: Use controller.store.now ?
        old_start = datetime.datetime.utcnow().replace(microsecond=0)
        offset = datetime.timedelta(hours=4)
        for i in range(number_of_facts):
            start = old_start + offset
            facts.append(fact_factory(start=start))
            old_start = start
        return facts

    return get_list_of_facts


# ***

FAKER_WORDS_NB = 6

FAKER_TABLE_LEN = 3


@pytest.fixture
def headers(faker):
    return faker.words(nb=FAKER_WORDS_NB)


@pytest.fixture
def row(faker):
    return faker.words(nb=FAKER_WORDS_NB)


@pytest.fixture
def table(faker):
    table = []
    for iters in range(FAKER_TABLE_LEN):
        table.append(faker.words(nb=FAKER_WORDS_NB))
    return table



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import copy

import pytest

from nark.items.tag import Tag


class TestTag(object):
    @classmethod
    def as_repr(cls, tag):
        if tag is None:
            return repr(tag)
        repred = (
            "Tag("
            "_name={name}, deleted={deleted}, freq={freq}, hidden={hidden}, pk={pk}"
            ")".format(
                pk=repr(tag.pk),
                name=repr(tag.name),
                deleted=repr(tag.deleted),
                freq=repr(tag.freq),
                hidden=repr(tag.hidden),
            )
        )
        return repred

    def test_init_valid(
        self, name_string_valid_parametrized_tag, pk_valid_parametrized
    ):
        """Make sure that Tag constructor accepts all valid values."""
        tag = Tag(name_string_valid_parametrized_tag, pk_valid_parametrized)
        assert tag.name == name_string_valid_parametrized_tag
        assert tag.pk == pk_valid_parametrized

    def test_init_invalid(self, name_string_invalid_parametrized_tag):
        """Make sure that Tag constructor rejects all invalid values."""
        with pytest.raises(ValueError):
            Tag(name_string_invalid_parametrized_tag)

    def test_as_tuple_include_pk(self, tag):
        """Make sure tags tuple representation works as intended and pk is included."""
        deleted = False
        hidden = False
        assert tag.as_tuple() == (tag.pk, tag.name, deleted, hidden)

    def test_as_tuple_exclude_pf(self, tag):
        """Make sure tags tuple representation works as intended and pk is excluded."""
        deleted = False
        hidden = False
        assert tag.as_tuple(include_pk=False) == (False, tag.name, deleted, hidden)

    def test_equal_fields_true(self, tag):
        """Make sure that two tags that differ only in their PK compare equal."""
        other_tag = copy.deepcopy(tag)
        other_tag.pk = 1
        assert tag.equal_fields(other_tag)

    def test_equal_fields_false(self, tag):
        """Make sure that two tags that differ not only in their PK compare unequal."""
        other_tag = copy.deepcopy(tag)
        other_tag.pk = 1
        other_tag.name += "foobar"
        assert tag.equal_fields(other_tag) is False

    def test__eq__false(self, tag):
        """Make sure that two distinct tags return ``False``."""
        other_tag = copy.deepcopy(tag)
        other_tag.pk = 1
        assert tag is not other_tag
        assert tag != other_tag

    def test__eq__true(self, tag):
        """Make sure that two identical categories return ``True``."""
        other_tag = copy.deepcopy(tag)
        assert tag is not other_tag
        assert tag == other_tag

    def test_is_hashable(self, tag):
        """Test that ``Tag`` instances are hashable."""
        assert hash(tag)

    def test_hash_method(self, tag):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(tag) == hash(tag.as_tuple())

    def test_hash_different_between_instances(self, tag_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Tag.as_tuple`` but for reassurance we test it anyway.
        """
        # 2023-12-19: Huh, this failed once on CI, but not reproducible.
        # - I wonder what the odds are that tag_factory() returns the
        #   same-named tag twice in a row...
        try:
            assert hash(tag_factory()) != hash(tag_factory())
        except AssertionError:
            # Per comment above, try again.
            assert hash(tag_factory()) != hash(tag_factory())

    def test__str__(self, tag):
        """Test string representation."""
        assert "{name}".format(name=tag.name) == str(tag)

    def test__repr__(self, tag):
        """Test representation method."""
        result = repr(tag)
        assert isinstance(result, str)
        expectation = TestTag.as_repr(tag)
        assert result == expectation



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Fixtures that are of general use."""

import datetime

import pytest

# *** (lb): Unused.


def convert_time_to_datetime(time_string):
    """
    Helper method.

    If given a %H:%M string, return a datetime.datetime object with todays
    date.
    """
    return datetime.datetime.combine(
        # MAYBE: Use controller.store.now ?
        datetime.datetime.utcnow().date(),
        datetime.datetime.strptime(time_string, "%H:%M").time(),
    )


# *** (lb): Unused.


@pytest.fixture
def raw_fact_with_persistent_activity(persistent_activity):
    """A raw fact whichs 'activity' is already present in the db."""
    return (
        "12:00-14:14 {a.name}@{a.category.name}".format(a=persistent_activity),
        {
            "start": convert_time_to_datetime("12:00"),
            "end": convert_time_to_datetime("14:14"),
            "activity": persistent_activity.name,
            "category": persistent_activity.category.name,
            "description": None,
        },
    )



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Base aggregate item fetch implementation."""

from sqlalchemy import func
from sqlalchemy.sql.expression import and_, or_

from ..objects import AlchemyActivity, AlchemyCategory, AlchemyFact, AlchemyTag
from . import (
    query_apply_limit_offset,
    query_prepare_datetime,
    query_sort_order_at_index,
)

__all__ = ("GatherBaseAlchemyManager",)


class GatherBaseAlchemyManager(object):
    """Base class aggregate query implementation for BaseAlchemyManager."""

    def __init__(self, *args, **kwargs):
        super(GatherBaseAlchemyManager, self).__init__(*args, **kwargs)

    # ***

    def gather(self, query_terms):
        """
        Returns matching items from the data store; and stats, if requested.

        Args:
            query_terms (nark.managers.query_terms.QueryTerms, required):
                The QueryTerms object defines the query settings used to find
                matching items, and it also defines how the results should be
                packaged. See the QueryTerms class for details.

        Returns:
            list: A list of matching item instances or (item, *statistics) tuples.
            - If raw results are requested, each item is the Alchemy<Item> object
              from SQLAlchemy (e.g., nark.backends.sqlalchemy.objects.AlchemyActivity).
              Otherwise, by default, each item is a first-class <Item> instance,
              hydrated from the Alchemy object (e.g., nark.items.activity.Activity).
            - If include_stats is requested, each result is a tuple consisting of
              the item as the first element, followed by additional calculated
              values (aggregates) pertaining to the query.
            - The list of results will be ordered according to the QueryTerms
              sort_cols and sort_orders options, if possible. Otherwise, for
              some queries that concatenate and post-process a value that is to
              be sorted on, the sorting is left as an exercise for the caller
              (because often the caller must do some post-processing of its own,
              so this method does not waste the cycles using, say, a subquery to
              do the sort in an outer query).
        """
        qt = query_terms

        # Get the Alchemy class, e.g., AlchemyActivity.
        alchemy_cls = self._gather_query_alchemy_cls

        # Compute the aggregate values if the user wants them returned with
        # the results, or if the user wants to sort the results accordingly.
        compute_usage = qt.include_stats or qt.sort_cols_has_any(
            "start", "usage", "time"
        )
        # If user is requesting filtering or sorting according to time, join Fact.
        requires_fact_table = self._gather_query_requires_fact(qt, compute_usage)

        def _gather_items():
            self.store.logger.debug(qt)

            query, agg_cols = _gather_query_start()

            query = self.query_filter_by_fact_times(
                query,
                qt.since,
                qt.until,
                qt.endless,
                qt.partial,
            )

            query = self.query_filter_by_activities(query, qt)

            query = self.query_filter_by_categories(query, qt)

            # (lb): We could add the match_tags option, but it'd be tricky
            # (we'd need to join on Facts, etc.), and does not seem of much
            # utility.
            #  query = self.query_filter_by_tags(query, qt)

            query = query_filter_by_search_term(query)

            query = self.query_filter_by_item_pk(query, alchemy_cls, qt.key)

            # FIXME/2020-06-03: Activity.deleted should not be used/useful.
            # (lb): And I've got some deleted = 0 and some deleted = 1 in my
            # database, but mostly deleted IS NULL, so skip deleted in WHERE
            # for attributes. (Remove this when removing deleted/hidden.)
            #  from . import query_apply_true_or_not
            #  query = query_apply_true_or_not(query, alchemy_cls.deleted, qt.deleted)

            query = query_group_by_aggregate(query, agg_cols)

            has_facts = requires_fact_table
            query = self.query_order_by_sort_cols(query, qt, has_facts, *agg_cols)

            query = query_apply_limit_offset(query, qt.limit, qt.offset)

            query = query_select_with_entities(query, agg_cols)

            self.query_prepared_trace(query)

            if qt.count_results:
                results = query.count()
            else:
                results = query.all()
                results = _gather_process_results(results)

            return results

        # ***

        def _gather_query_start():
            agg_cols = []

            if not requires_fact_table:
                query = self._gather_query_start_timeless(qt, alchemy_cls)
            else:
                if qt.include_stats or qt.sort_cols_has_any("usage"):
                    # (lb): I tried the COUNT() on the pk, e.g.,
                    #   count_col = func.count(alchemy_cls.pk).label('uses')
                    # but then if alchemy_cls is NULL (e.g., if Fact has an
                    # Activity where Category IS NULL), the count is 0, (e.g.,
                    # because COUNT(categories.id)). So do a broad COUNT(*), and
                    # then NULL rows will be counted (e.g., try `dob usage cat`).
                    # Just be sure there's a GROUP BY, otherwise the COUNT(*)
                    # will collapse all rows.
                    count_col = func.count().label("uses")
                    agg_cols.append(count_col)
                if qt.include_stats or qt.sort_cols_has_any("time"):
                    time_col = func.sum(
                        func.julianday(AlchemyFact.end)
                        - func.julianday(AlchemyFact.start)
                    ).label("span")
                    agg_cols.append(time_col)
                query = self._gather_query_start_aggregate(qt, agg_cols)

            return query, agg_cols

        # ***

        def query_filter_by_search_term(query):
            if not qt.search_terms:
                return query

            filters = []
            for term in qt.search_terms:
                filters.append(alchemy_cls.name.ilike("%{}%".format(term)))
            query = query.filter(or_(*filters))

            return query

        # ***

        def query_group_by_aggregate(query, agg_cols):
            if not agg_cols and not requires_fact_table:
                return query
            query = query.group_by(alchemy_cls.pk)
            return query

        # ***

        def query_select_with_entities(query, agg_cols):
            if not agg_cols and not requires_fact_table:
                return query
            # (lb): The aggregate query SELECTs all Fact columns, and it OUTER
            #  JOINs and GROUPs BY activities and categories or tags to produce
            #  the counts. But the Fact is meaningless after the group-by, as it
            #  represents a grouping. And we're not after the Fact; we want the
            #  attribute item (Activity, Category, or Tag). So we use with_entities
            #  to tell SQLAlchemy which columns to select -- it'll transform the
            #  query so that the SELECT fetches all the Activity columns; but the
            #  JOIN and GROUP BY remain the same. (Hey hey it's magic.)
            query = query.with_entities(alchemy_cls, *agg_cols)
            return query

        # ***

        def _gather_process_results(records):
            return self.query_process_results(
                records,
                raw=qt.raw,
                include_stats=compute_usage,
                requested_usage=qt.include_stats,
            )

        # ***

        return _gather_items()

    # ***

    def _gather_query_start_timeless(self, qt, alchemy_cls):
        # Query on, e.g., AlchemyActivity.
        query = self.store.session.query(alchemy_cls)
        return query

    def _gather_query_start_aggregate(self, qt, agg_cols):
        raise NotImplementedError

    # ***

    def _gather_query_requires_fact(self, qt, compute_usage):
        requires_fact_table = compute_usage or qt.since or qt.until or qt.endless
        return requires_fact_table

    # ***

    def query_filter_by_item_pk(self, query, alchemy_cls, key):
        if key is None:
            return query
        return query.filter(alchemy_cls.pk == key)

    # ***

    def query_filter_by_fact_times(
        self,
        query,
        since=None,
        until=None,
        endless=False,
        partial=False,
    ):
        def _query_filter_by_fact_times(query, since, until, endless, partial):
            fmt_since = query_prepare_datetime(since) if since else None
            fmt_until = query_prepare_datetime(until) if until else None
            if partial:
                query = _get_partial_overlaps(query, fmt_since, fmt_until)
            else:
                query = _get_complete_overlaps(query, fmt_since, fmt_until, endless)
            return query

        def _get_partial_overlaps(query, since, until):
            """Return all facts where either start or end falls within the timeframe."""
            if since and not until:
                # (lb): Checking AlchemyFact.end >= since is sorta redundant,
                # because AlchemyFact.start >= since should guarantee that.
                query = query.filter(
                    or_(
                        func.datetime(AlchemyFact.start) >= since,
                        func.datetime(AlchemyFact.end) >= since,
                    ),
                )
            elif not since and until:
                # (lb): Checking AlchemyFact.start <= until is sorta redundant,
                # because AlchemyFact.end <= until should guarantee that.
                # - Except maybe for an Active Fact?
                query = query.filter(
                    or_(
                        func.datetime(AlchemyFact.start) <= until,
                        func.datetime(AlchemyFact.end) <= until,
                    ),
                )
            elif since and until:
                query = query.filter(
                    or_(
                        and_(
                            func.datetime(AlchemyFact.start) >= since,
                            func.datetime(AlchemyFact.start) <= until,
                        ),
                        and_(
                            func.datetime(AlchemyFact.end) >= since,
                            func.datetime(AlchemyFact.end) <= until,
                        ),
                    )
                )
            else:
                pass
            return query

        def _get_complete_overlaps(query, since, until, endless=False):
            """Return all facts with start and end within the timeframe."""
            if since:
                query = query.filter(func.datetime(AlchemyFact.start) >= since)
            if until:
                query = query.filter(func.datetime(AlchemyFact.end) <= until)
            elif endless:
                query = query.filter(AlchemyFact.end == None)  # noqa: E711
            return query

        return _query_filter_by_fact_times(query, since, until, endless, partial)

    # ***

    def query_filter_by_activities(self, query, qt):
        query, criteria = self.query_criteria_filter_by_activities(query, qt)
        if criteria:
            query = query.filter(or_(*criteria))
        return query

    def query_criteria_filter_by_activities(self, query, qt):
        criteria = []
        for activity in qt.match_activities or []:
            criterion = self.query_filter_by_activity(activity)
            criteria.append(criterion)
        return query, criteria

    def query_filter_by_activity(self, activity):
        if activity is not None:
            activity_name = self.query_filter_by_activity_name(activity)
            if activity_name is None:
                # "Activity name must not be None."
                # (lb): This seems like a tedious branch to test and support. Same
                # goes for `activity is None` (the last else branch of this method).
                # Furthermore, the CLI does not let the user query for Activity
                # where name is None. Though at least the QueryTerms support
                # setting Activity=None (which is the last else branch of the
                # outer if-block). But as far as passing an Activity object to this
                # method, there's no production code that does that; you'd only
                # get here from a new test. Or maybe for some reason wiring
                # this path for some new feature.
                criterion = AlchemyActivity.pk == activity.pk
            else:
                # NOTE: Strict name matching, case and exactness.
                #       Not, say, func.lower(name) == func.lower(...),
                #       or using sqlalchemy ilike().
                criterion = AlchemyActivity.name == activity_name
        else:
            # activity is None.
            # (lb): Note that there's no production path that'll bring execution here.
            # - MAYBE: See preceding long comment: Add test; maybe wire from CLI.
            #   But what's the use case? You can find nameless Activities with `-a ''`.
            #   And there shouldn't be any Facts where Activity is NONE, right?
            #   Only unnamed Activities, and there's only at most one of those.
            criterion = AlchemyFact.activity == None  # noqa: E711
        return criterion

    def query_filter_by_activity_name(self, activity):
        activity_name = None
        try:
            if not activity.pk:
                activity_name = activity.name
        except AttributeError:
            activity_name = activity
        return activity_name

    # ***

    def query_filter_by_categories(self, query, qt):
        query, criteria = self.query_criteria_filter_by_categories(query, qt)
        if criteria:
            query = query.filter(or_(*criteria))
        return query

    def query_criteria_filter_by_categories(self, query, qt):
        criteria = []
        for category in qt.match_categories or []:
            criterion = self.query_filter_by_category(category)
            criteria.append(criterion)
        return query, criteria

    def query_filter_by_category(self, category):
        if category is not None:
            category_name = self.query_filter_by_category_name(category)
            if category_name is None:
                # See comment in query_filter_by_activity: this path not
                # reachable via production code.
                criterion = AlchemyCategory.pk == category.pk
            else:
                # NOTE: Strict name matching. Case and exactness count.
                criterion = AlchemyCategory.name == category_name
        else:
            # (lb): I tried to avoid delinting (the noqa) using is_, e.g.,
            #   criterion = (AlchemyActivity.category.is_(None))
            # but didn't happen:
            #   *** NotImplementedError: <function is_ at 0x7f75030f0280>
            criterion = AlchemyActivity.category == None  # noqa: E711
        return criterion

    def query_filter_by_category_name(self, category):
        category_name = None
        try:
            if not category.pk:
                category_name = category.name
        except AttributeError:
            category_name = category
        return category_name

    # ***

    def query_filter_by_tags(self, query, qt):
        # Note this method call only for Facts.get_all, but not other items' get_all.
        criteria = self.query_criteria_filter_by_tags(qt)
        query = query.filter(or_(*criteria))
        return query

    def query_criteria_filter_by_tags(self, qt):
        criteria = []
        for tag in qt.match_tags or []:
            criterion = self.query_filter_by_tag(tag)
            criteria.append(criterion)
        return criteria

    def query_filter_by_tag(self, tag):
        if tag is not None:
            tag_name = self.query_filter_by_tag_name(tag)
            if tag_name is None:
                # See comment in query_filter_by_activity: this path not
                # reachable via production code.
                criterion = AlchemyTag.pk == tag.pk
            else:
                criterion = AlchemyTag.name == tag_name
        else:
            # tag is None.
            criterion = AlchemyFact.tags == None  # noqa: E711
        return criterion

    def query_filter_by_tag_name(self, tag):
        tag_name = None
        try:
            if not tag.pk:
                tag_name = tag.name
        except AttributeError:
            tag_name = tag
        return tag_name

    # ***

    def query_order_by_sort_cols(self, query, query_terms, has_facts, *agg_cols):
        for idx, sort_col in enumerate(query_terms.sort_cols or []):
            direction = query_sort_order_at_index(query_terms.sort_orders, idx)
            query = self.query_order_by_sort_col(
                query,
                query_terms,
                sort_col,
                direction,
                has_facts,
                *agg_cols,
            )
        return query

    def query_order_by_sort_col(
        self,
        query,
        query_terms,
        sort_col,
        direction,
        has_facts,
        # The following columns are specific to Activity, Category, and Tag
        # gather() calls. The FactManager.gather overrides query_order_by_sort_col
        # to pass its own specific query columns.
        count_col=None,
        time_col=None,
    ):
        # Get the 'name' sort column, e.g., 'activity', 'category', or 'tag'.
        name_col = self._gather_query_order_by_name_col
        return self.query_usage_order_by(
            query,
            sort_col,
            direction,
            has_facts,
            name_col=name_col,
            count_col=count_col,
            time_col=time_col,
        )

    def query_usage_order_by(
        self,
        query,
        sort_col,
        direction,
        has_facts,
        name_col,
        count_col=None,
        time_col=None,
    ):
        # The query builder is responsible for ensuring that columns necessary
        # for the specified sort are included in the query. So this method can
        # assume that the necessary table (such as Fact, for order_by_start)
        # or aggregate columns (such as count_col or time_col) are available
        # when required for the sort.
        order_cols = []
        check_agg = False
        if sort_col == "start":
            assert has_facts
            order_cols = self.cols_order_by_start(query)
        elif sort_col == "usage":
            order_cols = [count_col]
            check_agg = True
        elif sort_col == "time":
            order_cols = [time_col]
            check_agg = True
        elif sort_col == "activity" or (
            name_col == "activity" and (sort_col == "name" or not sort_col)
        ):
            order_cols = [AlchemyActivity.name]
        elif sort_col == "category" or (
            name_col == "category" and (sort_col == "name" or not sort_col)
        ):
            order_cols = [AlchemyCategory.name]
        elif sort_col == "tag" or (
            name_col == "tag" and (sort_col == "name" or not sort_col)
        ):
            if self._gather_query_order_by_name_col == "tag":
                order_cols = [AlchemyTag.name]
            else:
                # Print a warning if called on Activity.get_all or Category.get_all,
                # because we don't join the Tag table for either of those.
                check_agg = True

        if not order_cols and check_agg:
            self.store.logger.warning("Invalid sort_col: {}".format(sort_col))
        elif not order_cols:
            self.store.logger.warning("Unknown sort_col: {}".format(sort_col))
        else:
            query = self.query_order_by_cols(query, direction, order_cols)
        return query

    # ***

    def cols_order_by_start(self, query):
        # Include end so that momentaneous Facts are sorted properly.
        # - And add PK, too, so momentaneous Facts are sorted predictably.
        return [AlchemyFact.start, AlchemyFact.end, AlchemyFact.pk]

    def query_order_by_start(self, query, direction):
        order_cols = self.cols_order_by_start(query)
        query = self.query_order_by_cols(query, direction, order_cols)
        return query

    def query_order_by_cols(self, query, direction, order_cols):
        directional_cols = [direction(col) for col in order_cols]
        query = query.order_by(*directional_cols)
        return query

    # ***

    def query_prepared_trace(self, query):
        if self.store.config["dev.catch_errors"]:
            # 2020-05-21: I don't generally like more noise in my tmux dev environment
            # logger pane, but I do like seeing the query, especially with all the
            # recent gather() development (improved grouping, sorting, and aggregates).
            logf = self.store.logger.warning
        else:
            logf = self.store.logger.debug
        logf("Query: {}".format(str(query)))

    # ***

    def query_process_results(
        self,
        records,
        raw,
        include_stats,
        requested_usage,
    ):
        def _query_process_results(records):
            if not records or not include_stats:
                return _process_records_items_only(records)
            return _process_records_items_and_aggs(records)

        def _process_records_items_only(records):
            if not raw:
                return [_as_hamster_or_none(item) for item in records]
            return records

        def _process_records_items_and_aggs(records):
            if not raw:
                return _process_records_items_and_aggs_hydrate(records)
            return records

        def _process_records_items_and_aggs_hydrate(records):
            if requested_usage:
                return [(_as_hamster_or_none(item), *cols) for item, *cols in records]
            return [_as_hamster_or_none(item) for item, *cols in records]

        def _as_hamster_or_none(item):
            # If query used outer join, and if, say, a Fact has an Activity set NULL,
            # or if an Activity has a Category set NULL, return None for the item.
            if item is not None:
                return item.as_hamster(self.store)
            return None

        return _query_process_results(records)

    # ***


# ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from nark.helpers.format_text import format_value_truncate


class TestFormatText(object):
    def test_format_value_truncate_one_word(self):
        truncated = format_value_truncate("abcdefghijkl", trunc_width=6)
        assert truncated == "..."

    def test_format_value_truncate_two_words(self):
        truncated = format_value_truncate("abc defghijkl", trunc_width=9)
        assert truncated == "abc..."

    def test_format_value_truncate_abc_de_6_plain(self):
        truncated = format_value_truncate(
            "abc de fghijkl",
            trunc_width=9,
        )
        assert truncated == "abc de..."

    def test_format_value_truncate_abc_de_6_italic_reset(self):
        truncated = format_value_truncate(
            "\x1b[3mabc de fghijkl\x1b[0m",
            trunc_width=9,
        )
        assert truncated == "\x1b[3mabc de...\x1b[0m"



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""This module provides nark raw fact parsing-related functions."""

import re
from datetime import timedelta
from gettext import gettext as _
from string import punctuation

import lazy_import

from .fact_time import (
    RE_PATTERN_RELATIVE_CLOCK,
    RE_PATTERN_RELATIVE_DELTA,
    datetime_from_clock_prior,
    parse_clock_time,
)
from .parse_errors import ParserInvalidDatetimeException

# Profiling: `import dateparser` takes ~ 0.2 seconds.
dateparser = lazy_import.lazy_module("dateparser")

# Profiling: load iso8601: ~ 0.004 secs.
iso8601 = lazy_import.lazy_module("iso8601")

__all__ = (
    "HamsterTimeSpec",
    "parse_dated",
    "parse_datetime_iso8601",
    "parse_relative_minutes",
)


# =================
# Notes on ISO 8601
# =================
#
# ASIDE: (lb): It's not my intent to write a datetime parser (there are plenty
# out there!), but our flexible factoid format lets the user specify datetimes
# in different, non-standard ways. E.g., the user can use relative time, which
# needs business logic to transform into a real datetime. So it's up to us to
# at least parse the datetime well enough to identify what type of format it's
# in, and then to either process it ourselves, or filter it through ``iso8601``
# or ``dateparser``.
#
# Some examples of ISO 8601 compliant datetimes
# ---------------------------------------------
#
#     ✓ 2018-05-14
#     ✓ 2018-05-14T22:29:24.123456+00:00
#     ✓ 2018-05-14T22:29:24+00:00
#     ✓ 2018-05-14T22:29:24Z
#     ✓ 20180514T222924Z
#
# Not all formats are supported by the ``iso8601`` parser
# -------------------------------------------------------
#
#     ✗ 2018-W20
#     ✗ 2018-W20-1
#     ✗ --05-14
#     ✗ 2018-134
#     ✓ 2018-12
#     ✗ 201805
#     ✓ 2018
#   __ ______________________________________________
#     ^ indicates if parse-worthy by iso8601 (✓ or ✗).
#
# And the ``iso8601`` parser also supports an extended format
# -----------------------------------------------------------
#
#   - The iso8601 parser allows ' ' in lieu of 'T'.
#
# The iso8601 parser format is: ``Date [Time [Timezone]]``
# --------------------------------------------------------
#
#   - Date and time are separated by ' ' or 'T'.
#
#   - Timezone immediately follow Time (no delimiter/space).
#
#   - Dates: YYYY-MM-DD | YYYYMMDD | YYYY-MM | YYYY
#
#   - Times: hh:mm:ss.nn | hhmmss.nn | hh:mm | hhmm | hh
#
#   - Time zones: <nothing> | Z | +/-hh:mm | +/-hhmm | +/-hh
#
#   - You can specify days or months without leading 0s [(lb): but why?].


class HamsterTimeSpec(object):
    """"""

    RE_HAMSTER_TIME = None

    def __init__(self):
        """Not implemented: Use class as static/global, not instantiated."""
        raise NotImplementedError

    @staticmethod
    def discern(hamster_time):
        """
        Check for magic datetimes:
          - '+/-n' relative;
          - 'nn:nn' clocktime;
          - ISO 8601 datetime.

        NOTE: This fcn. does not make datetime.datetime's; that's up to the caller.
        """
        dt, type_dt, sep, rest = None, None, None, None

        if HamsterTimeSpec.RE_HAMSTER_TIME is None:
            HamsterTimeSpec.setup_re()

        match = HamsterTimeSpec.RE_HAMSTER_TIME.match(hamster_time)
        if match is not None:
            say_what = match.groupdict()
            if say_what["relative"]:
                assert dt is None
                dt = say_what["relative"]
                type_dt = "relative"
            if say_what["clock_time"]:
                assert dt is None
                dt = say_what["clock_time"]
                type_dt = "clock_time"
            if say_what["datetime"]:
                assert dt is None
                dt = say_what["datetime"]
                type_dt = "datetime"
            assert dt is not None
            sep = say_what["sep"]
            rest = say_what["rest"]

        return dt, type_dt, sep, rest

    @staticmethod
    def setup_re():
        # NOTE: This pattern isn't perfect; and that's why we use the
        #       iso8601.parse_date routine.
        #
        #       (lb): It's because we use the ()? optionals.
        #       If one of the optionals is formatted incorrectly,
        #       the pattern here happily ignores it, because ?
        #       For instance, this matches, but the microseconds has an error:
        #
        #           RE_HAMSTER_TIME.match('2018-05-14 22:29:24.123x456+00:02')

        # Perhaps a lesser known feature: Hamster allows relative time!!
        pattern_relative = r"(?P<relative>([-+]?(\d+h)|[-+](\d+h)?\d+m?))"

        # Reminder: (?:...) is a non-capturing group.

        # Note: This regex does not verify hours and minutes in range 0..59.
        # - Whatever matches this, we'll re-parse with parse_clock_time, which
        #   does a more thorough match.
        pattern_just_clock = (
            "(?P<clock_time>"
            r"(?:(\d{1,2}:\d{2}:\d{2}|\d{1,2}:?\d{2}))"
            "(?:(a|am|A|AM|p|pm|P|PM)?)"
            ")"
        )

        # (lb): Treat 4 digits as clock time, not year, i.e.,
        #   `2030` should be 10:30 PM, not Jan 01, 2030.
        # This steals colon-less clock times:
        #   '(?:(\d{8}|\d{4}|\d{4}-\d{1,2}(-\d{1,2})?))'
        pattern_date = r"(?:(\d{8}|\d{4}-\d{1,2}(-\d{1,2})?))"
        # BEWARE: Does not verify hours and minutes in range 0..59.
        pattern_time = (
            # (lb): We could allow 3-digit times... but, no.
            #   '(?:\d{1,2})'
            r"(?:\d{2})"
            r"(?::?\d{2}"
            r"(?::?\d{2}"  # noqa: E131
            r"(?:\.\d+)?"  # noqa: E131
            ")?"
            ")?"
        )
        pattern_zone = (
            "(?:("
            "Z"  # noqa: E131
            "|"
            r"[+-]\d{2}(:?\d{2})?"
            "))?"
        )
        pattern_datetime = "(?P<datetime>{}([ T]{}{})?)".format(
            pattern_date, pattern_time, pattern_zone
        )

        hamster_pattern = r"(^|\s)({}|{}|{})(?P<sep>[,:]?)(?=\s|$)(?P<rest>.*)".format(
            pattern_relative,
            pattern_just_clock,
            pattern_datetime,
        )

        # Use re.DOTALL to match newlines, which might be part
        # of the <rest> of the factoid.
        HamsterTimeSpec.RE_HAMSTER_TIME = re.compile(hamster_pattern, re.DOTALL)

    @staticmethod
    def has_time_of_day(raw_dt):
        # Assuming format is year-mo-day separated from time of day by space or 'T'.
        parts = re.split(r" |T", raw_dt)
        if len(parts) != 2:
            return False
        # BEWARE: RE_PATTERN_RELATIVE_CLOCK does not validate range, e.g., 0..59.
        # - But this is just an assert, so should not fire anyway.
        assert re.match(RE_PATTERN_RELATIVE_CLOCK, parts[1]) is not None
        return True


# ***


def parse_dated(dated, time_now, cruftless=False):
    """"""

    def _parse_dated():
        if not isinstance(dated, str):
            # Let BaseFactManager.get_all() process, if not already datetime.
            return dated
        dt, type_dt, sep, rest = HamsterTimeSpec.discern(dated)
        if cruftless and rest:
            msg = _("Found more than datetime in")
            plus_sep = sep and " + ‘{}’".format(sep) or ""
            raise ParserInvalidDatetimeException(
                "{} “{}”: ‘{}’{} + ‘{}’".format(msg, dated, str(dt), plus_sep, rest)
            )
        parsed_dt = None
        if dt is not None:
            parsed_dt = datetime_from_discerned(dated, dt, type_dt)
        if parsed_dt is None:
            # FIXME/2020-05-09: Time zone awareness. Use local_tz.
            parsed_dt = parse_datetime_human(dated, time_now, local_tz=None)
        if parsed_dt is None:
            raise ParserInvalidDatetimeException(
                "{}: “{}”".format(_("Unparseable datetime"), dated)
            )
        return parsed_dt

    def datetime_from_discerned(dated, dt, type_dt):
        if type_dt == "datetime":
            # FIXME: (lb): Implement timezone/local_tz.
            dt_suss = parse_datetime_iso8601(dt, must=True, local_tz=None)
        # else, relative time, or clock time; let caller handle.
        elif type_dt == "clock_time":
            # Note that HamsterTimeSpec.discern is a little lazy and does
            # not verify the clock time is valid values, e.g., hours and
            # minutes between 0..59. But parse_clock_time cares.
            clock_time = parse_clock_time(dt)
            if not clock_time:
                return None

            # FIXME/2020-05-09 13:08: Use day_start here? (Or otherwise address issue
            # where clock time near now gets set nearly 24 hours later, not what user
            # meant.
            #            day_start = self.config['time.day_start']

            dt_suss = datetime_from_clock_prior(time_now, clock_time)
        else:
            assert type_dt == "relative"
            rel_mins, negative = parse_relative_minutes(dt)
            dt_suss = time_now + timedelta(minutes=rel_mins)
        return dt_suss

    return _parse_dated()


# ***

# The powerful dateparser parser can be a little too inclusive.
# - For instance, it'll accept `18 55`, parsed as the 18th day of today's month
#   in the year 2055; or even `44`, read as last century's 44th year, e.g., 1944.
# - The previous two examples, `18 55` and `44`, can be avoided by setting
#   STRICT_PARSING to True.
# - But strict parsing excludes such friendly single-day or single-month
#   inputs, e.g., `Monday` and `January`.
# - But strict parsing also allows input we'd like to ignore, e.g., even if
#   strict, `18:555` parses to Datetime(2055, 1, 8, 5, 0). What a stretch!
# - Note that a parser higher up in nark will have gone the quick route and
#   parsed the value already as a datetime if it was all numbers and maybe
#   some punctuation, e.g., something of the format `YYYY-MM-DD` or `HH:MM`
#   or similar. It's not relying on the dateparser for these simple queries.
# - So do an upfront check and skip the friendly parser if the input is
#   only numbers, whitespace, and/or punctuation, but no alpha characters.

RE_ONLY_09_WH_AND_PUNCT = re.compile(r"^[0-9\s{}]+$".format(re.escape(punctuation)))


def parse_datetime_get_settings(time_now=None, local_tz=None):
    # For settings def, see:
    #
    #   https://github.com/scrapinghub/dateparser/blob/master/docs/usage.rst#settings
    #
    #   dateparser/dateparser_data/settings.py
    #
    # Settings help:
    #
    # - PREFER_DATES_FROM:         defaults to current_period.
    #                              Options are future or past.
    # - SUPPORT_BEFORE_COMMON_ERA: defaults to False.
    # - PREFER_DAY_OF_MONTH:       defaults to current.
    #                              Could be first and last day of month.
    # - SKIP_TOKENS:               defaults to [‘t’]. Can be any string.
    # - TIMEZONE:                  defaults to UTC. Can be timezone abbrev
    #                              or any of tz database name as given here:
    #       https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    # - RETURN_AS_TIMEZONE_AWARE:  return tz aware datetime objects in
    #                              case timezone is detected in the date string.
    # - RELATIVE_BASE:             count relative date from this base date.
    #                              Should be datetime object.
    #
    # Settings defaults:
    #
    #   'PREFER_DATES_FROM': 'current_period',  # Or: 'past', 'future'.
    #   'PREFER_DAY_OF_MONTH': 'current',  # Or: 'first', 'last'.
    #   'SKIP_TOKENS': ["t"],
    #   'SKIP_TOKENS_PARSER': ["t", "year", "hour", "minute"],
    #   'TIMEZONE': 'local',
    #   'TO_TIMEZONE': False,
    #   'RETURN_AS_TIMEZONE_AWARE': 'default',
    #   'NORMALIZE': True,
    #   'RELATIVE_BASE': False,
    #   'DATE_ORDER': 'MDY',
    #   'PREFER_LOCALE_DATE_ORDER': True,
    #   'FUZZY': False,
    #   'STRICT_PARSING': False,
    #   'RETURN_TIME_AS_PERIOD': False,
    #   'PARSERS': default_parsers,
    #
    settings = {
        # Default to no time zone. We'll set True next, if local_tz.
        "RETURN_AS_TIMEZONE_AWARE": False,
        # So that, e.g., `dob list --since January` uses January 1st, and
        # not January whatever-today's-day-of-the-month-is, set DOM pref.
        "PREFER_DAY_OF_MONTH": "first",
        # Note that when STRICT_PARSING is False, it allows, e.g., `18 55`
        # to mean Datetime(2055, 12, 18, 0, 0), given a
        # today of Datetime(2015, 12, 10, 12, 30). But we
        # use RE_ONLY_09_WH_AND_PUNCT to avoid that.
        "STRICT_PARSING": False,
        # 2020-06-21: (lb): Code does not currently know time zone, and dateparser
        # says 'TIMEZONE' defaults to 'UTC', but I find I have to still specify
        # this value here, otherwise, when testing locally (but does not affect
        # Travic CI, apparently), the get_date_data() return is being adjusted
        # for my timezone. E.g., with a @freeze_time('2015-12-25 18:00'), calling
        # get_date_data('yesterday') returns '2015-12-24 12:00', which is 1 day
        # ago, but adjusted another 6 hours for 'America/Chicago'. IDGI, but
        # apparently being explicit here fixes the issue.
        "TIMEZONE": "UTC",
        # FEAT_REQ/2021-02-07: Should probably make date format configurable:
        # 'DATE_ORDER': 'MDY'|'DMY'|'YMD'|'YDM'
    }

    if time_now:
        # Using RELATIVE_BASE allows friendlies relative to other time, e.g.,
        # '1 hour ago', as in `dob from 1 hour ago to 2018-05-22 20:47` or
        # `dob list --since '1 hour ago'`, etc.
        settings["RELATIVE_BASE"] = time_now

    if local_tz:
        # NOTE: Uses machine-local tz, unless TIMEZONE set.
        settings["RETURN_AS_TIMEZONE_AWARE"] = True
        settings["TIMEZONE"] = local_tz

    return settings


def parse_datetime_human(datepart, time_now=None, local_tz=None):
    # Bouncer (aka Guard Clause): Callers have done their due diligence to
    # handler input that's numbers and punctuation only, so return early if
    # so.
    if RE_ONLY_09_WH_AND_PUNCT.match(datepart) is not None:
        return

    settings = parse_datetime_get_settings(time_now, local_tz)

    # Use the parse() wrapper class, so that the detected language is reused every
    # time, potentially speeding up a long import job. So avoid calling just this:
    #   parsed = dateparser.parse(datepart, settings=settings)
    ddp = dateparser.DateDataParser(settings=settings).get_date_data(datepart)
    # ddp is dict: 'date_obj' is None or the datetime;
    #              'period' is None or, e.g., 'day';
    #              'locale' is None or, e.g., 'en'.
    parsed = ddp["date_obj"]

    return parsed


# DateDataParser Notes
# ====================
#
# 2021-02-07: Note that dateparser uses non-numeric text as a language hint.
#
# Normally we don't care too much, because we're trying to be generous and
# let the user express themselves as loosely as possible. But the presence
# of text has some nuance that might be of interest, especially if you're
# trying to troubleshoot a peculiar parse result. (Which is why I ended
# up documenting this; normally the dob-import file-level parse_input.py
# parser strips a 'to' prefix, as in 'to YYYY-MM-DD', but it had been
# letting 'to' slip through if the line started with punctuation, e.g.,
# '[to YYYY-MM-DD...]', which led me to make the following observations.)
#
# For an example of how dateparser reacts to text, consider this setup:
#
#   (Pdb) ddp = dateparser.DateDataParser(settings=settings)
#
# Then, for example, a datetime string without text defaults to English:
#
#                          v (no prefix)
#   (Pdb) ddp.get_date_data('2021-02-05 22:04')
#   {'date_obj': datetime.datetime(2021, 2, 5, 22, 4), 'locale': 'en', ...}
#                                       ^^^^^ February 5   English^^
#
# But adding a seemingly innocuous 'o' prefix causes it to be interpreted
# as Polish, which defaults to Y-D-M, not Y-M-D, e.g.,
#
#                            v ('o')
#   (Pdb) ddp.get_date_data('o 2021-02-05 22:04')
#   {'date_obj': datetime.datetime(2021, 5, 2, 22, 4), 'locale': 'pl', ...}
#                                       ^^^^^ May 2        Polish ^^
#
# Going one character further, change 'o' to 'to' and it's considereds Finnish:
#
#                            vv ('to')
#   (Pdb) ddp.get_date_data('to 2021-02-05 22:04')
#   {'date_obj': datetime.datetime(2021, 5, 2, 22, 4), 'locale': 'fi', ...}
#                                       ^^^^^ May 2       Finnish ^^
#
# And then if the prefix is not a recognized article of speech from a supported
# language, it won't match at all, e.g.,
#
#   (Pdb) ddp.get_date_data('lo 2021-02-05 22:04')
#   {'date_obj': None, 'period': 'day', 'locale': None}
#
# Knowing that, we can probably guess the result if we indicate "the" date,
# mais *en français*:
#
#   (Pdb) ddp.get_date_data('le 2021-02-05 22:04')
#   {'date_obj': datetime.datetime(2021, 5, 2, 22, 4), 'locale': 'fr', ...}
#
# And finally here's what it looks like for English, using "the", and also
# including punctuation to show that DateDataParser ignores it:
#
#   (Pdb) ddp.get_date_data('*$%![the 2021-02-05 22:04')
#   {'date_obj': datetime.datetime(2021, 2, 5, 22, 4), 'locale': 'en', ...}


# ***


def parse_datetime_iso8601(datepart, must=False, local_tz=None):
    try:
        # NOTE: Defaults to datetime.timezone.utc.
        #       Uses naive if we set default_timezone=None.
        parsed = iso8601.parse_date(datepart, default_timezone=local_tz)
    except iso8601.iso8601.ParseError as err:
        parsed = None
        if must:
            raise ParserInvalidDatetimeException(
                _(
                    "Unable to parse iso8601 datetime: {} [{}]".format(
                        datepart, str(err)
                    )
                )
            )
    return parsed


# ***


def parse_relative_minutes(rel_time):
    rel_mins = None
    negative = None
    # NOTE: re.match checks for a match only at the beginning of the string.
    match = RE_PATTERN_RELATIVE_DELTA.match(rel_time)
    if match:
        parts = match.groupdict()
        rel_mins = 0
        if parts["minutes"]:
            rel_mins += int(parts["minutes"])
        if parts["hours"]:
            rel_mins += int(parts["hours"]) * 60
        if parts["signage"] == "-":
            negative = True  # Because there's no such thang as "-0".
            rel_mins *= -1
        else:
            negative = False
    return rel_mins, negative



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Query class module."""

from collections import namedtuple

QueryTermsTuple = namedtuple(
    "QueryTermsTuple",
    (
        "raw",
        "named_tuples",
        "include_stats",
        "count_results",
        "key",
        "since",
        "until",
        "endless",
        "exclude_ongoing",
        "partial",
        "deleted",
        "search_terms",
        "broad_match",
        "match_activities",
        "match_categories",
        "match_tags",
        "group_activity",
        "group_category",
        "group_tags",
        "group_days",
        "sort_cols",
        "sort_orders",
        "limit",
        "offset",
    ),
)


class QueryTerms(object):
    """
    The QueryTerms encapsulate all the input and output parameters of an item lookup.

    This class is a standalone class so that it's easy for frontends to use, too.
    """

    def __init__(self, **kwargs):
        self.setup_terms(**kwargs)

    def __str__(self):
        return " / ".join(
            [
                "raw?: {}".format(self.raw),
                "named?: {}".format(self.named_tuples),
                "stats?: {}".format(self.include_stats),
                "count?: {}".format(self.count_results),
                "key: {}".format(self.key),
                "since: {}".format(self.since),
                "until: {}".format(self.until),
                "endless: {}".format(self.endless),
                "excl-ongo?: {}".format(self.exclude_ongoing),
                "partial: {}".format(self.partial),
                "del?: {}".format(self.deleted),
                "terms: {}".format(self.search_terms),
                "broad: {}".format(self.broad_match),
                "acts: {}".format(self.match_activities),
                "cats: {}".format(self.match_categories),
                "tags: {}".format(self.match_tags),
                "grp-acts?: {}".format(self.group_activity),
                "grp-cats?: {}".format(self.group_category),
                "grp-tags?: {}".format(self.group_tags),
                "grp-days?: {}".format(self.group_days),
                "cols: {}".format(self.sort_cols),
                "ords: {}".format(self.sort_orders),
                "limit: {}".format(self.limit),
                "offset: {}".format(self.offset),
            ]
        )

    def setup_terms(
        self,
        raw=False,
        named_tuples=False,
        include_stats=None,
        count_results=False,
        key=None,
        since=None,
        until=None,
        # - SPIKE: (lb): Are 'endless' and 'exclude_ongoing' just opposites?
        endless=False,
        # - Use exclude_ongoing to omit the final, active Fact, if any,
        #   from the results.
        exclude_ongoing=None,
        partial=False,
        # The `deleted` option only applies to Facts.
        deleted=False,
        search_terms=None,
        broad_match=False,
        # - Note that item name matching is strict -- case and exactness count.
        match_activities=[],
        match_categories=[],
        match_tags=[],
        # - MEH: (lb): For parity, could add a 'tags' option to restrict the
        #   search to Activities used on Facts with specific 'tags', but how
        #   complicated and useless does that sound.
        # - Use the group_* flags to GROUP BY specific attributes.
        group_activity=False,
        group_category=False,
        group_tags=False,
        group_days=False,
        # - (lb): I added grouping support to FactManager.get_all via the options:
        #     group_activity
        #     group_category
        #     group_tags
        #   We could add them to this query, but it'd make it much more complex,
        #   and you'd get essentially the same results as using Fact.get_all (save
        #   for any Activities that are not applied to any Facts, but we can live
        #   with that gap in support). (tl;dr, use `dob list fact` or `dob usage fact`
        #   to group query results, and use the --column option if you want to tweak
        #   the output report columns, e.g., to match this method's output.)
        sort_cols=None,
        sort_orders=None,
        limit=None,
        offset=None,
    ):
        """
        Configures query parameters for item.get_all() and item.get_all_by_usage().

        Some of the settings affect the query, and some affect the returned results.

        Each of the query parameters is optional. Defaults are such that each
        argument default is falsey: it's either False, None, or an empty list.

        Args:
            raw: If True, returns 'raw' SQLAlchemy items (e.g., AlchemyFact).
                If False, returns first-class nark objects (e.g., Fact).
            named_tuples: If True, each result is a attribute-accessible tuple
                (like a namedtuple). If False, each result is a simple list. In
                either case, the first entry in always the item, and the order
                of additional details is always the same.
            include_stats: If True, computes additional details for each item or set
                of grouped items, and returns a list of tuples (with the item or
                aggregated item as the first element). Otherwise, if False, returns
                a list of matching items only. For Attribute, Category, and Tag
                queries, enable include_stats to receive a count of Facts that use
                the item, as well as the cumulative duration (end - start) of those
                Facts. For Facts, includes additional aggregate details.

            count_results: If True, return only a count of query matches (an integer).
                By default, count_results is False, and the method returns a list of
                results (of either items or tuples, depending on include_stats).

            key: If specified, look for an item with this PK. See also the get()
                method, if you do not need aggregate results.
            since: Restrict Facts to those that start at or after this time.
            until: Restrict Facts to those that end at or before this time.
                Note that a query will *not* match any Facts that start before and
                end after (e.g. that span more than) the specified timeframe.

            endless: If True, include the active Fact, if any, in the query.
            exclude_ongoing: If True, excldues the active Fact, in any.
            partial: If True, restrict Facts to those that start or end within the
                since-to-until time window.
            deleted: If True, include items marked 'deleted'.
            search_terms (None, or str list): Use to restrict to items whose name
                matches any on the specified search terms. Each comparison is case
                insensitive, and the match can occur in the middle of a string. If
                an item name matches one or more of the search terms, if any, it
                will be included in the results.
                * Use ``not`` before a search term to exclude its matches from the
                  results.
            broad_match: If True, apply search_terms fuzzy search to activity,
                category, and tag names.

            match_activities (list of items, each a nark.Activity, str, or None;
                optional): Matches only Activity(ies) or Facts that use the
                indicated Activity(ies) or Activity name(s) (exactly matched).
                The activity name can be specified as a string, or by passing a
                ``nark.Activity`` object whose name will be used. To match Facts
                without an Activity assigned, use ``None``. If ``match_activities``
                contains more than one item, Activities that exactly match or
                Facts that use *any* of the specified Activities will be included.
            match_categories (list of items, each a nark.Category, str, or None;
                optional): Matches only Category(ies) or Facts that use Activities
                that use the indicated Category(ies) or Category name(s) (exactly
                matched). The category name can be specified as a string, or by
                passing a ``nark.Category`` object whose name will be used. To match
                Facts with an Activity that does not have a Category assigned, use
                ``None``. If ``match_categories`` contains more than one item,
                Categories that exactly match or Facts that use an Activity with
                *any* of the specified Categories will be included.
            match_tags (list of items, each a nark.Tag, str, or None; optional):
                Matchies only Tag(s) or Facts that use the indicated Tag(s) or
                Tag name(s) (exactly matched). The tag name can be specified as
                a string, or by passing a ``nark.Tag`` object whose name will be
                used. To match Facts without a Tag assigned, use ``tag=None``.
                If ``match_tags`` contains more than one item, Tags that exactly
                match or Facts that use an Tag with *any* of the specified Tags
                will be included.

            group_activity: If True, GROUP BY the Activity name, unless group_category
                is also True, then GROUP BY the Activity PK and the Category PK.
            group_category: If True, GROUP BY the Category PK.
            group_tags: If True, group by the Tag PK.
            group_days: If True, group by the Fact start date (e.g., 1999-12-31,
                i.e., truncating clock time).

            sort_cols (str list, optional): Which column(s) to sort by.
                - If not aggregating results, defaults to 'name' and orders
                  by item name.
                - When aggregating results (include_stats=True) or searching
                  Facts, defaults to 'start', and orders results by Fact start.
                - Choices include: 'start', 'time', 'day', 'name', 'activity,
                  'category', 'tag', 'usage', and 'fact'.
                - Note that 'start' and 'usage' only apply if include_stats,
                  and 'day' is only valid when group_days is True.
            sort_orders (str list, optional): Specifies the direction of each
                sort specified by sort_cols. Use the string 'asc' or 'desc'
                in the corresponding index of sort_orders that you want applied
                to the corresponding entry in soft_cols. If there is no
                corresponding entry in sort_orders for a specific sort_cols
                entry, that sort column is ordered in ascending order.

            limit (int, optional): Query "limit".
            offset (int, optional): Query "offset".
        """
        self.raw = raw
        self.named_tuples = named_tuples
        self.include_stats = include_stats

        self.count_results = count_results

        self.key = key
        self.since = since
        self.until = until
        self.endless = endless
        self.exclude_ongoing = exclude_ongoing
        self.partial = partial
        self.deleted = deleted

        self.search_terms = search_terms
        self.broad_match = broad_match

        self.match_activities = match_activities
        self.match_categories = match_categories
        self.match_tags = match_tags

        self.group_activity = group_activity
        self.group_category = group_category
        self.group_tags = group_tags
        self.group_days = group_days

        self.sort_cols = sort_cols
        self.sort_orders = sort_orders

        self.limit = limit
        self.offset = offset

    # ***

    def as_tuple(self):
        return QueryTermsTuple(
            raw=self.raw,
            named_tuples=self.named_tuples,
            include_stats=self.include_stats,
            count_results=self.count_results,
            key=self.key,
            since=self.since,
            until=self.until,
            endless=self.endless,
            exclude_ongoing=self.exclude_ongoing,
            partial=self.partial,
            deleted=self.deleted,
            search_terms=self.search_terms,
            broad_match=self.broad_match,
            match_activities=self.match_activities,
            match_categories=self.match_categories,
            match_tags=self.match_tags,
            group_activity=self.group_activity,
            group_category=self.group_category,
            group_tags=self.group_tags,
            group_days=self.group_days,
            sort_cols=self.sort_cols,
            sort_orders=self.sort_orders,
            limit=self.limit,
            offset=self.offset,
        )

    def __eq__(self, other):
        if other is not None and not isinstance(other, QueryTermsTuple):
            other = other.as_tuple()
        return self.as_tuple() == other

    # ***

    @property
    def is_grouped(self):
        is_grouped = (
            self.group_activity
            or self.group_category
            or self.group_tags
            or self.group_days
        )
        return is_grouped

    def sort_cols_has_any(self, *args):
        return set(self.sort_cols or []).intersection(set(args))

    @property
    def sorts_cols_has_stat(self):
        return self.sort_cols_has_any("usage", "time", "day")



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Base fixtures available to all nark tests."""

import pytest

from nark.control import NarkControl
from nark.manager import BaseStore

# (lb): I might be doing this wrong, but works: So that downstream packages can
# reuse (DRY) nark fixtures, such fixtures are defined in the public package.
# - Note that I tried pytest_plugins, but didn't work, e.g.,
#     pytest_plugins = ('nark.tests', 'nark.tests.conftest', ...)
#   then:
#     $ py.test
#     ...
#     fixture 'base_config' not found
# - So we *-glob import the fixtures instead.
#   - This isn't the cleanest solution, but it's also not uncommon:
#     - conftest.py itself is already magic: py.test loads this module for
#       every test_{}.py file, and everything herein is exported.
#     - The pytest-factoryboy register() functions also magically inject
#       fixtures into the namespace (that themselves are mutated from a
#       corresponding classname, e.g., FactFactory becomes a fixture named
#       fact_factory).
#     - So don't bark too loudly about the opaque import-all.
# - Linter ref:
#   - F401 'nark.tests.conftest.*' imported but unused
#   - F403 'from nark.tests.conftest import *' used; unable to detect undefined names
from nark.tests.conftest import *  # noqa: F401, F403

pytest_plugins = (
    "tests.config.envvar_prefix",
    "tests.config.init_app_dirs",
)


@pytest.fixture
def controller(base_config):
    """Provide a basic controller."""
    # From hamster-lib: "[TODO] Parametrize over all available stores."
    # (lb): And yet in dob there's still just the one for SQLite.
    controller = NarkControl(base_config)
    yield controller
    controller.store.cleanup()


@pytest.fixture
def basestore(base_config):
    """Provide a generic ``storage.BaseStore`` instance using ``baseconfig``."""
    store = BaseStore(base_config)
    return store



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import xml


class TestXMLWriter(object):
    """Make sure the XML writer works as expected."""

    def test_xml_writer_start_document(self, xml_writer, faker):
        """Make sure a XML main document and a facts list child element is set up."""
        ename = faker.word()
        xml_writer.start_document(ename)
        assert xml_writer.document
        assert xml_writer.fact_list
        assert xml_writer.fact_list.tagName == ename

    def test_xml_writer__write_fact(self, xml_writer, fact, mocker):
        """Make sure that the attributes attached to the fact matche our
        expectations."""
        # (lb): I tried patching the Class method so we could call write_facts, e.g.,
        #   from xml.dom.minidom import Node
        #   mocker.patch.object(Node, 'appendChild')
        #   xml_writer.write_facts([fact])
        # But then the elem.setAttribute() methods do not work.
        # So call _write_fact directly, and test write_facts elsewhere.
        xml_writer.start_document("facts")
        mocker.patch.object(xml_writer.fact_list, "appendChild")
        xml_writer._write_fact(idx=0, fact=fact)
        # Grab the document element created and passed to appendChild.
        result = xml_writer.fact_list.appendChild.call_args[0][0]
        fact_start = fact.start_fmt(xml_writer.datetime_format)
        fact_end = fact.end_fmt(xml_writer.datetime_format)
        fact_duration = fact.format_delta(style=xml_writer.duration_fmt)
        assert result.getAttribute("start") == fact_start
        assert result.getAttribute("end") == fact_end
        assert result.getAttribute("duration") == fact_duration
        assert result.getAttribute("activity") == fact.activity_name
        assert result.getAttribute("category") == fact.category_name
        assert result.getAttribute("description") == fact.description_or_empty

    def test_xml_writer_write_report(self, xml_writer, mocker, headers, row):
        xml_writer.start_document("results")
        mocker.patch.object(xml_writer.fact_list, "appendChild")
        xml_writer._write_result(row, headers)
        result = xml_writer.fact_list.appendChild.call_args[0][0]
        # DUNNO/2023-12-19: CI: Sometimes fails (or at least has failed once).
        # - E.g., AssertionError: assert 'special' == 'interview'
        # - Looking at _write_result, I'm not sure how this could happen.
        for idx, col in enumerate(headers):
            assert result.getAttribute(col) == row[idx]

    def test_xml_writer_write_facts__close(self, xml_writer, fact, path):
        """Make sure the calendar is actually written do disk before file is closed."""
        xml_writer.write_facts([fact])
        with open(path, "r", encoding="utf8") as fobj:
            result = xml.dom.minidom.parse(fobj)
            assert result.toxml()

    def test_xml_writer_write_report__close(self, xml_writer, table, headers):
        """Make sure the calendar is actually written do disk before file is closed."""
        output_path = xml_writer.output_file.name
        xml_writer.write_report(table, headers)
        with open(output_path, "r") as fobj:
            result = xml.dom.minidom.parse(fobj)
            assert result.toxml()



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""``nark`` storage object managers."""

from gettext import gettext as _

from sqlalchemy.exc import IntegrityError

from .gather_base import GatherBaseAlchemyManager

__all__ = ("BaseAlchemyManager",)


class BaseAlchemyManager(GatherBaseAlchemyManager):
    """Base class for sqlalchemy managers."""

    def __init__(self, *args, **kwargs):
        super(BaseAlchemyManager, self).__init__(*args, **kwargs)

    # ***

    def add_and_commit(self, alchemy_item, raw=False, skip_commit=False):
        """
        Adds the item to the data store, and perhaps calls commit.

        Generally, unless importing Facts, the session is committed
        after an item is added or updated. However, when adding or
        updating a Fact, we might also create other items (activity,
        category, tags), so we delay committing until everything is
        added/updated.
        """

        def _add_and_commit():
            session_add()
            session_commit_maybe()
            result = prepare_item()
            self.store.logger.debug(_("Added item: {!r}".format(result)))
            return result

        def session_add():
            self.store.session.add(alchemy_item)

        def session_commit_maybe():
            if skip_commit:
                return
            try:
                self.store.session.commit()
            except IntegrityError as err:
                message = _(
                    "An error occured! Are you sure that the {0}'s name "
                    "or ID is not already present? Error: '{1}'.".format(
                        self.__class__.__name__,
                        err,
                    )
                )
                self.store.logger.error(message)
                raise ValueError(message)

        def prepare_item():
            result = alchemy_item
            if not raw:
                result = alchemy_item.as_hamster(self.store)
            return result

        return _add_and_commit()

    # ***

    def get_all(self, query_terms=None, **kwargs):
        """Returns matching items from the data store; and stats, if requested.

        get_all() is similar to get_all_by_usage(), but get_all() prefers not to
        include usage statistics in the results, and prefers to sort by item name.

        See gather method and QueryTerms class docstrings for more details.
        """
        query_terms, kwargs = self._gather_prepare_query_terms(query_terms, **kwargs)
        if query_terms.include_stats is None:
            query_terms.include_stats = False
        if query_terms.sort_cols is None:
            query_terms.sort_cols = ("name",)
        return super(BaseAlchemyManager, self).get_all(query_terms, **kwargs)

    def get_all_by_usage(self, query_terms=None, **kwargs):
        """Returns matching items from the data store; and stats, if requested.

        get_all_by_usage() is an alias to get_all() that prefers to enable
        include_stats, and to sort by the usage calculation.

        See gather method and QueryTerms class docstrings for more details.
        """
        query_terms, kwargs = self._gather_prepare_query_terms(query_terms, **kwargs)
        if query_terms.include_stats is None:
            query_terms.include_stats = True
        if query_terms.sort_cols is None:
            query_terms.sort_cols = ("usage",)
        return super(BaseAlchemyManager, self).get_all(query_terms, **kwargs)

    # ***


# ***



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Tests for nark.backends.sqlalchemy.managers.gather_base.GatherBaseAlchemyManager."""

import pytest

from nark.backends.sqlalchemy.managers.gather_base import GatherBaseAlchemyManager
from nark.backends.sqlalchemy.objects import AlchemyActivity
from nark.items.activity import Activity
from nark.managers.query_terms import QueryTerms

# (lb): You'll find much of GatherBaseAlchemyManager tested by derived classes' tests.
# What's not tested by one of the derived classes should be covered here, albeit often
# we use the Activity as the search class.


class TestGatherBaseAlchemyManager:
    """"""

    # MAYBE: Move five_report_facts_ctl from dob to nark, which is a session-scoped
    # five-Facts fixture, which can noticeably speed up tests. For now, use the
    # simpler set_of_alchemy_facts which isn't the quickest fixture, but we can
    # just not perform these tests for the other item types (category and tag)
    # because the code under test here is not item-specific.

    def test_get_all_count_results(self, alchemy_store, set_of_alchemy_facts):
        count = alchemy_store.activities.get_all(count_results=True)
        assert count == len(set_of_alchemy_facts)

    def test_get_all_no_query_terms_raw(self, alchemy_store, set_of_alchemy_facts):
        """Test query_process_results._process_records_items_only second branch."""
        results = alchemy_store.activities.get_all(raw=True)
        # Each fixture Fact was assigned a unique Activity.
        assert len(results) == len(set_of_alchemy_facts)
        # Each result is a raw AlchemyActivity.
        assert isinstance(results[0], AlchemyActivity)

    def test_get_all_include_stats_act(self, alchemy_store, set_of_alchemy_facts):
        results = alchemy_store.activities.get_all(include_stats=True)
        # Each fixture Fact was assigned a unique Activity.
        assert len(results) == len(set_of_alchemy_facts)
        activity, uses, duration = results[0]
        # Each result's first item is a hydrated Activity.
        assert isinstance(activity, Activity)
        # Because all fixture Activity names are unique, only 1 use of it.
        assert uses == 1

    def test_get_all_include_stats_raw(self, alchemy_store, set_of_alchemy_facts):
        """Test query_process_results._process_records_items_and_aggs second branch."""
        results = alchemy_store.activities.get_all(include_stats=True, raw=True)
        # Each fixture Fact was assigned a unique Activity.
        assert len(results) == len(set_of_alchemy_facts)
        activity, uses, duration = results[0]
        # Each result's first item is a raw AlchemyActivity.
        assert isinstance(activity, AlchemyActivity)
        # Because all fixture Activity names are unique, only 1 use of it.
        assert uses == 1

    def test_get_all_include_stats_not_implemented(self, alchemy_store, mocker):
        """Test that unimplemented _gather_query_start_aggregate raises."""
        unclassed = GatherBaseAlchemyManager()
        unclassed._gather_query_alchemy_cls = mocker.MagicMock()
        unclassed.store = mocker.MagicMock()
        mocker.patch.object(unclassed, "_gather_query_alchemy_cls")
        qt = QueryTerms(include_stats=True)
        # Test _gather_query_start_aggregate raises.
        with pytest.raises(NotImplementedError):
            unclassed.gather(qt)

    def test_get_all_compute_stats_exclude(self, alchemy_store, set_of_alchemy_facts):
        """Test query_process_results._process_records_items_and_aggs_hydrate branch."""
        results = alchemy_store.activities.get_all(sort_cols=["usage"])
        # Each fixture Fact was assigned a unique Activity.
        assert len(results) == len(set_of_alchemy_facts)
        # Each result is a hydrated Activity.
        assert isinstance(results[0], Activity)

    def test_get_all_fact_without_activity(self, alchemy_store, alchemy_fact):
        """Test query_process_results._as_hamster_or_none second branch."""
        alchemy_fact.activity = None
        results = alchemy_store.activities.get_all(include_stats=True)
        activity, uses, duration = results[0]
        assert activity is None
        assert uses == 1

    def test_get_all_endless_true(self, alchemy_store, set_of_alchemy_facts_active):
        results = alchemy_store.activities.get_all(endless=True)
        # Only 1 Fact is endless.
        assert len(results) == 1

    def test_get_all_query_prepared_trace(self, alchemy_store, mocker):
        alchemy_store.config["dev.catch_errors"] = True
        mocker.patch.object(alchemy_store.logger, "warning")
        alchemy_store.activities.get_all()
        assert alchemy_store.logger.warning.called



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import copy

import pytest

from nark.items.activity import Activity

from .test_category import TestCategory


class TestActivity(object):
    @classmethod
    def as_repr(cls, activity):
        if activity is None:
            return repr(activity)
        repred = (
            "Activity(_name={name}, category={category}, "
            "deleted={deleted}, hidden={hidden}, pk={pk})"
        ).format(
            pk=repr(activity.pk),
            name=repr(activity.name),
            category=TestCategory.as_repr(activity.category),
            deleted=repr(activity.deleted),
            hidden=repr(activity.hidden),
        )
        return repred

    def test_init_valid(
        self,
        name_string_valid_parametrized,
        pk_valid_parametrized,
        category_valid_parametrized,
        deleted_valid_parametrized,
    ):
        """Test that init accepts all valid values."""
        activity = Activity(
            name_string_valid_parametrized,
            pk=pk_valid_parametrized,
            category=category_valid_parametrized,
            deleted=deleted_valid_parametrized,
        )
        assert activity.name == name_string_valid_parametrized
        assert activity.pk == pk_valid_parametrized
        assert activity.category == category_valid_parametrized
        assert activity.deleted == bool(deleted_valid_parametrized)

    def test_init_invalid(self, name_string_invalid_parametrized):
        """
        Test that init rejects all invalid values.

        Note:
            Right now, the only aspect that can have concievable invalid value
            is the name.
        """
        with pytest.raises(ValueError):
            Activity(name_string_invalid_parametrized)

    def test_create_from_composite(self, activity):
        result = Activity.create_from_composite(activity.name, activity.category.name)
        assert result.name == activity.name
        assert result.category == activity.category

    def test_as_tuple_include_pk(self, activity):
        """Make sure that conversion to a tuple matches our expectations."""
        expecting = (
            activity.pk,
            activity.name,
            (
                activity.category.pk,
                activity.category.name,
                activity.category.deleted,
                activity.category.hidden,
            ),
            activity.deleted,
            activity.hidden,
        )
        assert activity.as_tuple() == expecting

    def test_as_tuple_exclude_pk(self, activity):
        """Make sure that conversion to a tuple matches our expectations."""
        expecting = (
            False,
            activity.name,
            (
                False,
                activity.category.name,
                activity.category.deleted,
                activity.category.hidden,
            ),
            activity.deleted,
            activity.hidden,
        )
        assert activity.as_tuple(include_pk=False) == expecting

    def test_equal_fields_true(self, activity):
        """
        Make sure that two activities that differ only in their PK compare equal.
        """
        other = copy.deepcopy(activity)
        other.pk = 1
        assert activity.equal_fields(other)

    def test_equal_fields_false(self, activity):
        """
        Make sure that two activities that differ not only in their PK
        compare unequal.
        """
        other = copy.deepcopy(activity)
        other.pk = 1
        other.name += "foobar"
        assert activity.equal_fields(other) is False

    def test__eq__false(self, activity):
        """Make sure that two distinct activities return ``False``."""
        other = copy.deepcopy(activity)
        other.pk = 1
        assert activity is not other
        assert activity != other

    def test__eq__true(self, activity):
        """Make sure that two identical activities return ``True``."""
        other = copy.deepcopy(activity)
        assert activity is not other
        assert activity == other

    def test_is_hashable(self, activity):
        """Test that ``Category`` instances are hashable."""
        assert hash(activity)

    def test_hash_method(self, activity):
        """Test that ``__hash__`` returns the hash expected."""
        assert hash(activity) == hash(activity.as_tuple())

    def test_hash_different_between_instances(self, activity_factory):
        """
        Test that different instances have different hashes.

        This is actually unneeded as we are merely testing the builtin ``hash``
        function and ``Category.as_tuple`` but for reassurance we test it anyway.
        """
        assert hash(activity_factory()) != hash(activity_factory())

    def test__str__without_category(self, activity):
        activity.category = None
        assert str(activity) == "{name}".format(name=activity.name)

    def test__str__with_category(self, activity):
        assert str(activity) == "{name} ({category})".format(
            name=activity.name, category=activity.category.name
        )

    def test__repr__with_category(self, activity):
        """Make sure our debugging representation matches our expectations."""
        result = repr(activity)
        assert isinstance(result, str)
        expectation = TestActivity.as_repr(activity)
        assert result == expectation

    def test__repr__without_category(self, activity):
        """Make sure our debugging representation matches our expectations."""
        activity.category = None
        result = repr(activity)
        assert isinstance(result, str)
        expectation = TestActivity.as_repr(activity)
        assert result == expectation



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os

import lazy_import

from nark.backends.sqlalchemy import objects

from ....helpers.legacy_db import upgrade_legacy_db_hamster_applet
from ....managers.migrate import BaseMigrationsManager

# Profiling: Loading `migrate` takes ~ 0.090 seconds.
migrate_exceptions = lazy_import.lazy_module(
    "sqlalchemy_migrate_hotoffthehamster.exceptions"
)
migrate_versioning_cfgparse = lazy_import.lazy_module(
    "sqlalchemy_migrate_hotoffthehamster.versioning.cfgparse"
)
migrate_versioning_api = lazy_import.lazy_module(
    "sqlalchemy_migrate_hotoffthehamster.versioning.api"
)

__all__ = ("MigrationsManager",)


class MigrationsManager(BaseMigrationsManager):
    """"""

    @property
    def basepath(self):
        return self.migrations_path()

    @property
    def config(self):
        return self.migrations_config()

    # ***

    # - Prefer using an existing SQLAlchemy Engine, rather than opening the URL.
    # - The migrate class accepts either a database URL, or an existing Engine
    #   object. Given a path, the migrate_versioning package will just open a
    #   new connection to the URL, which works fine for files on disk or actual
    #   URLs on the net. But for tests, especially those run via CliRunner where
    #   it's more difficult to mock or otherwise skip normal initialization code,
    #   use the existing Engine object.
    #   - This supports using an SQLite store in :memory:, for which there is no
    #     other way to connect to the same database other than through the Engine
    #     ((lb): At least, not that I know of; and makes sense there wouldn't be).
    # - Note that using CliRunner poses a challenge with passing variables, and
    #   test variables that need to be applied to the Click code under test
    #   generally need to be applied using a MagicMock or a PropertyMock. However,
    #   the Engine is already always available at a global location (at least the
    #   way we've got our code wired, which follows convention), so rather than
    #   mock it in, we just always look for the existing Engine, even in production.
    #   - The SQLAlchemy Engine object is stashed at objects.metadata.bind, and is
    #     meant to be "held globally for the lifetime of a single application...."
    #       https://docs.sqlalchemy.org/en/13/core/connections.html
    @property
    def engine_or_url(self):
        return objects.metadata.bind or self.store.db_url

    # ***

    def control(self, version=None, engine=None):
        """Mark a database as under version control."""
        current_ver = self.version()
        if current_ver is None:
            url = self.engine_or_url
            try:
                migrate_versioning_api.version_control(
                    url,
                    self.basepath,
                    version=version,
                    config=self.config,
                )
                return True
            except migrate_exceptions.DatabaseAlreadyControlledError:
                return False
        elif current_ver == 0:
            return False
        else:
            return None

    def downgrade(self):
        """Downgrade the database according to its migration version."""
        current_ver = self.version()
        if current_ver is None:
            return None
        latest_ver = migrate_versioning_api.version(
            self.basepath,
            config=self.config,
        )
        if not latest_ver:
            return None
        assert current_ver <= latest_ver
        if current_ver > 0:
            next_version = current_ver - 1
            url = self.engine_or_url
            migrate_versioning_api.downgrade(
                url,
                self.basepath,
                version=next_version,
                config=self.config,
            )
            return True
        else:
            return False

    def upgrade(self):
        """Upgrade the database according to its migration version."""
        current_ver = self.version()
        if current_ver is None:
            return None
        latest_ver = migrate_versioning_api.version(
            self.basepath,
            config=self.config,
        )
        if not latest_ver:
            return None
        assert current_ver <= latest_ver
        if current_ver < latest_ver:
            next_version = current_ver + 1
            url = self.engine_or_url
            migrate_versioning_api.upgrade(
                url,
                self.basepath,
                version=next_version,
                config=self.config,
            )
            return True
        else:
            return False

    def version(self):
        """Returns the current migration of the active database."""
        url = self.engine_or_url
        try:
            return migrate_versioning_api.db_version(
                url,
                self.basepath,
                config=self.config,
            )
        except migrate_exceptions.DatabaseNotControlledError:
            return None

    def latest_version(self):
        """Returns the latest version defined by the application."""
        try:
            repo_latest = migrate_versioning_api.version(
                self.basepath,
                config=self.config,
            )
            return int(repo_latest.value)
        except migrate_exceptions.DatabaseNotControlledError:
            return None

    # ***

    def migrations_config(self):
        """Return sqlalchemy-migrate migrate.cfg a hot-n-ready config object."""
        # See nark/migrations/migrate.cfg.example for descriptions.
        config = migrate_versioning_cfgparse.Parser()
        config["db_settings"] = {}
        config["db_settings"]["repository_id"] = "nark-migrations"
        config["db_settings"]["version_table"] = "migrate_version"
        config["db_settings"]["required_dbs"] = "[]"
        config["db_settings"]["use_timestamp_numbering"] = "False"
        return config

    def migrations_path(self):
        # (lb): This is a little awkward. But there's not
        # another convenient way to do this, is there?
        path = os.path.abspath(
            os.path.join(
                # Meh. We could also do dirname(nark.__file__) and use fewer ..'s.
                #    path = os.path.join(os.path.dirname(nark.__file__), 'migrations')
                # except that nark is not a package, it's this package!
                os.path.dirname(__file__),
                "../../../migrations",
            )
        )
        return path

    # ***

    def legacy_upgrade_from_hamster_applet(self, db_path):
        upgrade_legacy_db_hamster_applet(db_path)

    def legacy_upgrade_from_hamster_lib(self):
        # (lb): I'm not sure how much traction hamster-lib had.
        #  So I'm not sure this function is needed.
        #  And hamster-libbers out there need a hand?
        #  You'll need to rename 3 things is all, I'm sure:
        #    Table renamed: facttags → fact_tags
        #    Cols renamed: facts.start_time/end_time → facts.start/end
        #  Please open a ticket if you really want this feature!
        raise NotImplementedError



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import csv
from gettext import gettext as _

import pytest


class TestPlaintextWriter(object):
    def test_plaintext_writer_init(self, plaintext_writer):
        """Make sure that initialition provides us with a ``csv.writer`` instance."""
        assert plaintext_writer.csv_writer
        assert plaintext_writer.csv_writer.dialect == csv.get_dialect("excel")

    def test_plaintext_writer_headers(self, path, plaintext_writer):
        """Make sure that initialition writes header as expected."""
        expectations = (
            "Start time",
            "End time",
            "Duration",
            "Activity",
            "Category",
            "Description",
            "Deleted",
        )
        plaintext_writer.write_facts(facts=[])
        with open(plaintext_writer.output_file.name, "r") as fobj:
            reader = csv.reader(fobj, dialect="excel")
            header = next(reader)
        for field, expectation in zip(header, expectations):
            if isinstance(field, str):
                assert field == expectation
            else:
                assert field.decode("utf-8") == expectation

    def test_plaintext_writer_fact_as_tuple_no_category(self, plaintext_writer, fact):
        """Make sure that ``None`` category values translate to ``empty strings``."""
        fact.activity.category = None
        result = plaintext_writer.fact_as_tuple(fact)
        cat_idx = plaintext_writer.facts_headers().index(_("Category"))
        assert result[cat_idx] == ""

    def test_plaintext_writer_fact_as_tuple_with_category(self, plaintext_writer, fact):
        """Make sure that category references translate to their names."""
        result = plaintext_writer.fact_as_tuple(fact)
        cat_idx = plaintext_writer.facts_headers().index(_("Category"))
        assert result[cat_idx] == fact.category.name

    def test_plaintext_writer__write_fact(self, fact, plaintext_writer):
        """Make sure the writen fact is what we expect."""
        fact_tuple = plaintext_writer.fact_as_tuple(fact)
        plaintext_writer._write_fact(idx=0, fact=fact)
        plaintext_writer._close()
        with open(plaintext_writer.output_file.name, "r", encoding="utf8") as fobj:
            reader = csv.reader(fobj, dialect=plaintext_writer.dialect)
            # If we had called write_facts, would need to ignore headers,
            # e.g.,
            #   next(reader)
            # and then fetch the line.
            line = next(reader)
            for field, expectation in zip(line, fact_tuple):
                if isinstance(field, str):
                    assert field == expectation
                else:
                    assert field.decode("utf-8") == expectation

    def test_plaintext_writer_write_report(self, plaintext_writer, table, headers):
        plaintext_writer.write_report(table, headers)
        with open(plaintext_writer.output_file.name, "r") as fobj:
            reader = csv.reader(fobj, dialect=plaintext_writer.dialect)
            # The first line is the headers.
            csv_headers = next(reader)
            assert csv_headers == headers
            # Remaining lines are data rows.
            for row in table:
                assert row == next(reader)
            with pytest.raises(StopIteration):
                next(reader)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

# Most of nark/helpers/fact_time.py is tested incidentally by other tests.
# (lb): Here I'm just filling in the blanks.

import datetime

import pytest

from nark.helpers.fact_time import (
    datetime_from_clock_after,
    datetime_from_clock_prior,
    day_end_datetime,
    must_be_datetime_or_relative,
    must_not_start_after_end,
    parse_clock_time,
)


@pytest.fixture
def clock_time():
    clock_time = "22:59"
    return parse_clock_time(clock_time)


@pytest.fixture
def dt_relative_early():
    dt_relative = datetime.datetime(2015, 12, 10, 12, 30, 0)
    return dt_relative


@pytest.fixture
def dt_relative_later():
    dt_relative = datetime.datetime(2015, 12, 10, 23, 30, 0)
    return dt_relative


class TestFactTime(object):
    def test_datetime_from_clock_prior_day_before(self, dt_relative_early, clock_time):
        new_dt = datetime_from_clock_prior(dt_relative_early, clock_time)
        assert new_dt == datetime.datetime(2015, 12, 9, 22, 59)

    def test_datetime_from_clock_prior_same_day(self, dt_relative_later, clock_time):
        new_dt = datetime_from_clock_prior(dt_relative_later, clock_time)
        assert new_dt == datetime.datetime(2015, 12, 10, 22, 59, 0)

    def test_datetime_from_clock_after_same_day(self, dt_relative_early, clock_time):
        new_dt = datetime_from_clock_after(dt_relative_early, clock_time)
        assert new_dt == datetime.datetime(2015, 12, 10, 22, 59, 0)

    def test_datetime_from_clock_after_day_after(self, dt_relative_later, clock_time):
        new_dt = datetime_from_clock_after(dt_relative_later, clock_time)
        assert new_dt == datetime.datetime(2015, 12, 11, 22, 59, 0)

    def test_day_end_datetime_day_starts_midnight(self, dt_relative_early):
        end = day_end_datetime(end_date=dt_relative_early)
        assert end == datetime.datetime(2015, 12, 10, 23, 59, 59)

    def test_day_end_datetime_day_starts_special(self, dt_relative_early):
        start_time = datetime.time(hour=15, minute=34)
        end = day_end_datetime(dt_relative_early, start_time)
        assert end == datetime.datetime(2015, 12, 11, 15, 33, 59)

    def test_must_be_datetime_or_relative_datetime(self):
        dt = datetime.datetime(2015, 12, 10, 12, 30, 33, 1234)
        new_dt = must_be_datetime_or_relative(dt)
        assert new_dt == datetime.datetime(2015, 12, 10, 12, 30, 33)

    def test_must_be_datetime_or_relative_not_relative_raises(self):
        dt = "foo"
        with pytest.raises(TypeError) as excinfo:
            must_be_datetime_or_relative(dt)
        errmsg = "Expected time entry to indicate relative time, not:"
        assert str(excinfo.value).startswith(errmsg)

    def test_must_be_datetime_or_relative_not_relative_valid(self):
        dt = "-5h13m"
        new_dt = must_be_datetime_or_relative(dt)
        assert new_dt == dt

    def test_must_be_datetime_or_relative_not_relative_invalid(self):
        dt = object()
        with pytest.raises(TypeError) as excinfo:
            must_be_datetime_or_relative(dt)
        errmsg = "Time entry not `datetime`, relative string, or `None`, but:"
        assert str(excinfo.value).startswith(errmsg)

    def test_must_not_start_after_end_start_after_end_raises(
        self,
        dt_relative_early,
        dt_relative_later,
    ):
        range_tuple = (dt_relative_later, dt_relative_early)
        with pytest.raises(ValueError) as excinfo:
            must_not_start_after_end(range_tuple)
        errmsg = "Start after end!"
        assert str(excinfo.value) == errmsg

    def test_must_not_start_after_end_end_after_start_valid(
        self,
        dt_relative_early,
        dt_relative_later,
    ):
        range_tuple = (dt_relative_early, dt_relative_later)
        new_rt = must_not_start_after_end(range_tuple)
        assert new_rt == range_tuple

    def test_parse_clock_time_valid(self, clock_time):
        clock_time = "22:59"
        parsed_ct = parse_clock_time(clock_time)
        assert parsed_ct == (22, 59, 0)

    def test_parse_clock_time_invalid(self):
        not_clock_time = "24:99"
        parsed_ct = parse_clock_time(not_clock_time)
        assert parsed_ct is None



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""JSON writer output format module."""

import json

from . import ReportWriter

__all__ = ("JSONWriter",)


class JSONWriter(ReportWriter):
    def __init__(self):
        """Initialize a new JSONWriter instance."""
        super(JSONWriter, self).__init__()

    def fact_as_dict(self, fact):
        kvals = {
            "start": fact.start_fmt(self.datetime_format),
            "end": fact.end_fmt(self.datetime_format),
            "activity": fact.activity_name,
            "duration": fact.format_delta(style=self.duration_fmt),
            "category": fact.category_name,
            "description": fact.description_or_empty,
        }
        return kvals

    def init_result_list(self):
        self.result_list = []

    def write_result_list(self):
        json.dump(self.result_list, self.output_file)
        n_written = len(self.result_list)
        del self.result_list
        return n_written

    def write_facts_list(self, facts):
        self.init_result_list()
        super(JSONWriter, self).write_facts_list(facts)
        n_written = self.write_result_list()
        return n_written

    def _write_fact(self, idx, fact):
        kvals = self.fact_as_dict(fact)
        self.result_list.append(kvals)

    def write_report_table(self, table, headers, tabulation=None):
        self.init_result_list()
        super(JSONWriter, self).write_report_table(table, headers, tabulation)
        n_written = self.write_result_list()
        return n_written

    def _write_result(self, row, headers, tabulation=None):
        kvals = {header: value for header, value in zip(headers, row)}
        self.result_list.append(kvals)



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime
import os

import pytest

from nark.config import ConfigRoot, decorate_config


@pytest.fixture
def config_root():
    config_root = ConfigRoot
    config_root.forget_config_values()
    return config_root


class TestConfigObjToBackendConfig(object):
    """Make sure that conversion works expected."""

    def test_regular_usecase(self, configobj_instance):
        """Make sure basic mechanics work and int/time types are created."""
        configobj, expectation = configobj_instance
        result = decorate_config(configobj).as_dict(unmutated=True)
        assert result == expectation

    def test_decorate_config_on_config(self, config_root):
        """Test that decorate_config given existing config returns that config."""
        assert decorate_config(config_root) is config_root


class TestNarkConfigurableDb(object):
    """Test NarkConfigurableDb ConfigRoot.section."""

    def test_db_path_with_appdirs(self, tmpdir, mocker):
        path = tmpdir.strpath
        # Note that if not already loaded, we'd need to import the package:
        #   import easy_as_pypi_appdirs  # noqa: F401
        # but pytest_plugins includes init_app_dirs, which imports it.
        mocker.patch("easy_as_pypi_appdirs.appdirs.user_data_dir", return_value=path)
        # The ConfigRoot will already have been created without AppDirs setup,
        # so tell it to forget what it knows, and it'll scan defaults again.
        config_root = ConfigRoot
        config_root.forget_config_values()
        db_path = config_root["db"]["path"]
        assert db_path == os.path.join(path, "dob.sqlite")


class TestNarkConfigurableTime(object):
    """Test NarkConfigurableTime ConfigRoot.section."""

    # *** nark.config._strptime_day_start

    # These tests implicitly test nark.config._strptime_day_start.

    def test_day_start_empty_string(self, config_root):
        config_root.asobj.time.day_start.value = ""
        assert config_root["time"]["day_start"] == datetime.time(0, 0, 0)

    def test_day_start_datetime_time(self, config_root):
        time = datetime.time(23, 59, 44)
        config_root.asobj.time.day_start.value = time
        assert config_root["time"]["day_start"] == time

    def test_day_start_from_text_valid(self, config_root):
        time = "21:34:56"
        config_root.asobj.time.day_start.value = time
        assert config_root["time"]["day_start"] == datetime.time(21, 34, 56)

    def test_day_start_from_text_invalid(self, config_root):
        time = "25:66:67"
        with pytest.raises(ValueError) as excinfo:
            config_root.asobj.time.day_start.value = time
        assert str(excinfo.value).startswith("Unrecognized value for setting")

    # *** ephemeral settings

    def test_tz_aware_default(self, config_root):
        assert config_root["time"]["tz_aware"] is False

    def test_default_tzinfo_default(self, config_root):
        assert config_root["time"]["default_tzinfo"] == ""



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import datetime

import pytest
from freezegun import freeze_time

from nark.managers.query_terms import QueryTerms

# ***


class TestFactManager:
    def test_save_fact_endless_active(self, basestore, fact, mocker):
        """
        Make sure that passing a fact without end (aka 'endless, ongoing,
        active fact') triggers the correct method.
        """
        magic_fact = {}
        mocker.patch.object(basestore.facts, "_add", return_value=magic_fact)
        fact.end = None
        new_fact = basestore.facts.save(fact)
        assert basestore.facts._add.called
        assert new_fact is magic_fact

    def test_save_fact_too_brief_value_error(self, basestore, fact):
        """Ensure that a fact with too small of a time delta raises an exception."""
        delta = datetime.timedelta(
            seconds=(basestore.config["time.fact_min_delta"] - 1)
        )
        fact.end = fact.start + delta
        with pytest.raises(ValueError):
            basestore.facts.save(fact)

    def test_save_fact_no_fact_min_delta(self, basestore, fact, mocker):
        """Ensure that a fact with too small of a time delta raises an exception."""
        magic_fact = {}
        mocker.patch.object(basestore.facts, "_add", return_value=magic_fact)
        # Note that out config defined the type as int, so use 0, not None.
        basestore.config["time.fact_min_delta"] = 0
        fact.end = fact.start
        new_fact = basestore.facts.save(fact)
        assert basestore.facts._add.called
        assert new_fact is magic_fact

    def test_add_not_implemented(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts._add(fact)

    def test_update_not_implemented(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts._update(fact)

    def test_remove_not_implemented(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts.remove(fact)

    def test_get_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.get(12)

    def test_get_all_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            # Note that BaseManager.get_all actually *is* implemented, but
            # it calls self.gather(), which won't be implemented in basestore.
            # So this runs a little code (checking for and converting the 'since'
            # and 'until' arguments to datetime objects) before the raise.
            basestore.facts.get_all()

    def test_get_all_by_usage_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.get_all_by_usage()

    def test_gather_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.gather(query_terms=None)

    # *** get_all since/until argument tests.

    # (lb): I think these were from old hamster-lib where it made more sense
    # for these to be complicated. The code has been refactored so much that
    # these tests do a lot of work for not much code or unique branch coverage.

    @pytest.mark.parametrize(
        ("since", "until", "expectation"),
        [
            (None, None, {"since": None, "until": None}),
            # Various since info.
            (
                datetime.date(2014, 4, 1),
                None,
                {"since": datetime.datetime(2014, 4, 1, 5, 30, 0), "until": None},
            ),
            (
                datetime.time(13, 40, 25),
                None,
                {"since": datetime.datetime(2015, 4, 1, 13, 40, 25), "until": None},
            ),
            (
                datetime.datetime(2014, 4, 1, 13, 40, 25),
                None,
                {"since": datetime.datetime(2014, 4, 1, 13, 40, 25), "until": None},
            ),
            # Various until info.
            (
                None,
                datetime.date(2014, 2, 1),
                {"since": None, "until": datetime.datetime(2014, 2, 2, 5, 29, 59)},
            ),
            (
                None,
                datetime.time(13, 40, 25),
                {"since": None, "until": datetime.datetime(2015, 4, 1, 13, 40, 25)},
            ),
            (
                None,
                datetime.datetime(2014, 4, 1, 13, 40, 25),
                {"since": None, "until": datetime.datetime(2014, 4, 1, 13, 40, 25)},
            ),
        ],
    )
    @freeze_time("2015-04-01 18:00")
    def test_get_all_various_since_and_until_times(
        self,
        basestore,
        mocker,
        since,
        until,
        expectation,
    ):
        """Test that time conversion matches expectations."""
        mocker.patch.object(
            basestore.facts,
            "gather",
        )
        query_terms = QueryTerms(since=since, until=until)
        # MAYBE/2020-05-25: (lb): I don't quite like that get_all mutates query_terms.
        basestore.facts.get_all(query_terms)
        assert basestore.facts.gather.called
        actual_qt = basestore.facts.gather.call_args[0][0]
        expect_qt = QueryTerms(
            since=expectation["since"],
            until=expectation["until"],
        )
        assert actual_qt == expect_qt

    @pytest.mark.parametrize(
        ("since", "until"),
        [
            (
                datetime.date(2015, 4, 5),
                datetime.date(2012, 3, 4),
            ),
            (
                datetime.datetime(2015, 4, 5, 18, 0, 0),
                datetime.datetime(2012, 3, 4, 19, 0, 0),
            ),
        ],
    )
    def test_get_all_until_before_since(self, basestore, since, until):
        """Test that we throw an exception if passed until time is before since time."""
        with pytest.raises(ValueError):
            basestore.facts.get_all(since=since, until=until)

    @pytest.mark.parametrize(
        ("since", "until"),
        [
            # (lb): This test used to cause TypeError, because get_all used to not
            # parse the input, but expected datetime objects instead. So the string
            # values in the parameters here would trigger TypeError. But now the
            # parsing is part of get_all (because DRY), so now we test ValueError.
            # - Note that since > until here, so ValueError, and tests mixed types.
            (datetime.date(2015, 4, 5), "2012-03-04"),
            ("2015-04-05 18:00:00", "2012-03-04 19:00:00"),
        ],
    )
    def test_get_all_invalid_date_types(self, basestore, mocker, since, until):
        """Test that we throw an exception if we receive invalid date/time objects."""
        with pytest.raises(ValueError):
            basestore.facts.get_all(since=since, until=until)

    # ***

    @freeze_time("2015-10-03 14:45")
    def test_get_today(self, basestore, mocker):
        """Make sure that method uses appropriate timeframe."""
        mocker.patch.object(basestore.facts, "get_all", return_value=[])
        results = basestore.facts.get_today()
        assert results == []
        assert basestore.facts.get_all.call_args[1] == {
            "since": datetime.datetime(2015, 10, 3, 5, 30, 0),
            "until": datetime.datetime(2015, 10, 4, 5, 29, 59),
        }

    @freeze_time("2015-10-03 14:45")
    def test_day_end_datetime_no_end_date(self, basestore):
        """Make sure that method uses appropriate timeframe."""
        basestore.config["time.day_start"] = datetime.time(5, 30)
        until = basestore.facts.day_end_datetime(end_date=None)
        assert until == datetime.datetime(2015, 10, 4, 5, 29, 59)

    # *** stop_current_fact tests.

    @freeze_time("2019-02-01 18:00")
    @pytest.mark.parametrize(
        "hint",
        (
            None,
            datetime.timedelta(minutes=10),
            datetime.timedelta(minutes=300),
            datetime.timedelta(seconds=-10),
            datetime.timedelta(minutes=-10),
            datetime.datetime(2019, 2, 1, 19),
        ),
    )
    def test_stop_current_fact_with_hint(
        self,
        basestore,
        base_config,
        endless_fact,
        hint,
        mocker,
    ):
        """
        Make sure we can stop an 'ongoing fact' and that it will have an end set.

        Please note that ever so often it may happen that the factory generates
        a endless_fact with ``Fact.start`` after our mocked today-date. In order
        to avoid confusion the easies fix is to make sure the mock-today is well
        in the future.
        """
        now = datetime.datetime.now()  # the freeze_time time, above.
        # (lb): The FactFactory sets start to faker.Faker().date_time(),
        # which is not constrained in any way, and might set Fact.start in
        # the future, or close to now (freeze_time), in which case our 'hint'
        # value might otherwise cause the test to end the Fact before it
        # starts, eliciting the complaint:
        #   ValueError:
        #       Cannot end the Fact before it started.
        #       Try editing the Fact instead.
        # So ensure that setting end (relative to now) won't fail, by
        # possibly moving Fact.start. (lb): Does it make setting Fact.start
        # to a fake time worth it if we have to possibly change it anyway?
        # - Start with the fact_min_delta padding, which will also cause a
        #   complaint, if end isn't far enough ahead of start.
        max_delta_secs = base_config["time"]["fact_min_delta"]
        if hint and isinstance(hint, datetime.timedelta) and hint.total_seconds() < 0:
            # Subtract hint's seconds, which is a negative value, so really,
            # increase the max_delta_secs, because hint will be subtracted
            # from now later (expected_end = now + hint).
            max_delta_secs -= hint.total_seconds()
        max_start = None
        if max_delta_secs:
            max_start = now - datetime.timedelta(seconds=max_delta_secs)
        if endless_fact.start > max_start:
            endless_fact.start = max_start
        # NOTE: The `fact` fixture simply adds a second Fact to the db, after
        #       having added the endless_fact Fact.
        if hint:
            if isinstance(hint, datetime.datetime):
                expected_end = hint
            else:
                # Add hint to now, e.g., datetime.datetime(2019, 2, 1, 18) + hint.
                expected_end = now + hint
        else:
            # NOTE: Because freeze_time, datetime.now() === datetime.utcnow().
            expected_end = datetime.datetime.now().replace(microsecond=0)
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        mocker.patch.object(
            basestore.facts,
            "_add",
        )
        basestore.facts.stop_current_fact(hint)
        assert basestore.facts.endless.called
        assert basestore.facts._add.called
        fact_to_be_added = basestore.facts._add.call_args[0][0]
        assert fact_to_be_added.end == expected_end
        fact_to_be_added.end = None
        assert fact_to_be_added == endless_fact

    @freeze_time("2019-02-01 18:00")
    @pytest.mark.parametrize("hint", (datetime.datetime(2019, 2, 1, 17, 59),))
    def test_stop_current_fact_with_end_in_the_past(
        self,
        basestore,
        base_config,
        endless_fact,
        hint,
        mocker,
    ):
        """
        Make sure that stopping an 'ongoing fact' with end before start raises.
        """
        # Set start to the freeze_time time, above.
        endless_fact.start = datetime.datetime.now()
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        mocker.patch.object(
            basestore.facts,
            "_add",
        )
        with pytest.raises(ValueError):
            basestore.facts.stop_current_fact(hint)
        assert basestore.facts.endless.called
        assert not basestore.facts._add.called

    def test_stop_current_fact_invalid_offset_hint(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """
        Make sure that stopping with an offset hint that results in end > start
        raises an error.
        """
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        now = datetime.datetime.now().replace(microsecond=0)
        offset = (now - endless_fact.start).total_seconds() + 100
        offset = datetime.timedelta(seconds=-1 * offset)
        with pytest.raises(ValueError):
            basestore.facts.stop_current_fact(offset)

    def test_stop_current_fact_invalid_datetime_hint(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """
        Make sure that stopping with a datetime hint that results in end > start
        raises an error.
        """
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        with pytest.raises(ValueError):
            basestore.facts.stop_current_fact(
                endless_fact.start - datetime.timedelta(minutes=30),
            )

    def test_stop_current_fact_invalid_hint_type(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """Make sure that passing an invalid hint type raises an error."""
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        with pytest.raises(TypeError):
            basestore.facts.stop_current_fact(str())

    # ***

    # Note that get_current_fact() also gets tested by way of
    # stop_current_fact(), as well as find_latest_fact().

    def test_get_current_fact_datebase_integrity_issue_multiple_active_facts(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """Make sure that if 2 Facts are Active, get_current_fact fails."""
        two_endless = [endless_fact, endless_fact]
        mocker.patch.object(basestore.facts, "endless", return_value=two_endless)
        with pytest.raises(Exception):
            basestore.facts.get_current_fact()

    # *** find_latest_fact

    def test_find_latest_fact_finds_fact_ongoing_exists(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """Ensure find_latest_fact finds 'ongoing' Fact."""
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        restrict = "ongoing"
        found = basestore.facts.find_latest_fact(restrict=restrict)
        assert found is endless_fact

    def test_find_latest_fact_finds_fact_ongoing_not_found(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """Ensure find_latest_fact returns None when no 'ongoing' Fact."""
        mocker.patch.object(basestore.facts, "endless", return_value=[])
        restrict = "ongoing"
        found = basestore.facts.find_latest_fact(restrict=restrict)
        assert found is None

    def test_find_latest_fact_finds_fact_ongoing_too_many(
        self,
        basestore,
        endless_fact,
        mocker,
    ):
        """Ensure find_latest_fact raises when there is more than 1 'ongoing' Fact."""
        two_endless = [endless_fact, endless_fact]
        mocker.patch.object(basestore.facts, "endless", return_value=two_endless)
        restrict = "ongoing"
        with pytest.raises(Exception):
            basestore.facts.find_latest_fact(restrict=restrict)

    def test_find_latest_fact_finds_fact_ended(
        self,
        basestore,
        fact,
        mocker,
    ):
        """Ensure find_latest_fact find an 'ended' Fact."""
        mocker.patch.object(basestore.facts, "get_all", return_value=[fact])
        restrict = "ended"
        found = basestore.facts.find_latest_fact(restrict=restrict)
        assert found is fact

    # *** find_oldest_fact

    def test_find_oldest_fact_found(
        self,
        basestore,
        fact,
        mocker,
    ):
        """Ensure find_oldest_fact find a fact when get_all returns one."""
        mocker.patch.object(basestore.facts, "get_all", return_value=[fact])
        found = basestore.facts.find_oldest_fact()
        assert found is fact

    def test_find_oldest_fact_not_found(
        self,
        basestore,
        fact,
        mocker,
    ):
        """Ensure find_oldest_fact find a fact when get_all returns none."""
        mocker.patch.object(basestore.facts, "get_all", return_value=[])
        found = basestore.facts.find_oldest_fact()
        assert found is None

    # *** endless

    def test_get_endless_fact_with_ongoing_fact(
        self,
        basestore,
        endless_fact,
        fact,
        mocker,
    ):
        """Make sure we return the 'ongoing_fact'."""
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        fact = basestore.facts.endless()
        assert fact == fact
        assert fact is fact

    def test_get_endless_fact_without_ongoing_fact(self, basestore, mocker):
        """Make sure that we raise a KeyError if there is no 'ongoing fact'."""
        mocker.patch.object(basestore.facts, "endless", return_value=[])
        fact = basestore.facts.endless()
        assert fact == []

    # *** cancel_current_fact

    def test_cancel_current_fact(self, basestore, endless_fact, fact, mocker):
        """Make sure we return the 'ongoing_fact'."""
        mocker.patch.object(basestore.facts, "endless", return_value=[endless_fact])
        mocker.patch.object(
            basestore.facts,
            "remove",
        )
        result = basestore.facts.cancel_current_fact()
        assert basestore.facts.endless.called
        assert basestore.facts.remove.called
        assert result is endless_fact  # Because mocked.
        # FIXME: Where's the test that actually tests FactManager.endless()?

    def test_cancel_current_fact_without_endless_fact(self, basestore, mocker):
        """Make sure that we raise a KeyError if there is no 'ongoing fact'."""
        mocker.patch.object(basestore.facts, "endless", return_value=[])
        with pytest.raises(KeyError):
            basestore.facts.cancel_current_fact()
        assert basestore.facts.endless.called

    # ***

    def test_starting_at_not_implemented(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts.starting_at(fact)

    def test_ending_at_not_implemented(self, basestore, fact):
        with pytest.raises(NotImplementedError):
            basestore.facts.ending_at(fact)

    def test_antecedent_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.antecedent()

    def test_subsequent_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.subsequent()

    def test_strictly_during_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.strictly_during(start=None, end=None)

    def test_surrounding_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.surrounding(fact_time=None)

    def test_endless_not_implemented(self, basestore):
        with pytest.raises(NotImplementedError):
            basestore.facts.endless()



# This file exists within 'nark':
#
#   https://github.com/tallybark/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from ....managers.tag import BaseTagManager
from ..objects import (
    AlchemyActivity,
    AlchemyCategory,
    AlchemyFact,
    AlchemyTag,
    fact_tags,
)
from . import query_apply_true_or_not
from .manager_base import BaseAlchemyManager

__all__ = ("TagManager",)


class TagManager(BaseAlchemyManager, BaseTagManager):
    """ """

    def __init__(self, *args, **kwargs):
        super(TagManager, self).__init__(*args, **kwargs)

    # ***

    def get_or_create(self, tag, raw=False, skip_commit=False):
        """
        Custom version of the default method in order to provide access to
        alchemy instances.

        Args:
            tag (nark.Tag): Tag we want.
            raw (bool): Wether to return the AlchemyTag instead.

        Returns:
            nark.Tag or None: Tag.
        """
        self.store.logger.debug("Received: {!r} / raw: {}".format(tag, raw))

        try:
            tag = self.get_by_name(tag.name, raw=raw)
        except KeyError:
            tag = self._add(tag, raw=raw, skip_commit=skip_commit)
        return tag

    # ***

    def _add(self, tag, raw=False, skip_commit=False):
        """
        Add a new tag to the database.

        This method should not be used by any client code. Call ``save`` to make
        the decission wether to modify an existing entry or to add a new one is
        done correctly..

        Args:
            tag (nark.Tag): nark Tag instance.
            raw (bool): Wether to return the AlchemyTag instead.

        Returns:
            nark.Tag: Saved instance, as_hamster()

        Raises:
            ValueError: If the name to be added is already present in the db.
            ValueError: If tag passed already got an PK. Indicating that update
                would be more appropriate.
        """
        self.adding_item_must_not_have_pk(tag)

        alchemy_tag = AlchemyTag(
            pk=None,
            name=tag.name,
            deleted=bool(tag.deleted),
            hidden=bool(tag.hidden),
        )

        result = self.add_and_commit(
            alchemy_tag,
            raw=raw,
            skip_commit=skip_commit,
        )

        return result

    # ***

    def _update(self, tag):
        """
        Update a given Tag.

        Args:
            tag (nark.Tag): Tag to be updated.

        Returns:
            nark.Tag: Updated tag.

        Raises:
            ValueError: If the new name is already taken.
            ValueError: If tag passed does not have a PK.
            KeyError: If no tag with passed PK was found.
        """
        self.store.logger.debug("Received: {!r}".format(tag))

        if not tag.pk:
            message = _(
                "The Tag passed ('{!r}') does not have a PK."
                " We do not know which entry to modify."
            ).format(tag)
            self.store.logger.error(message)
            raise ValueError(message)
        alchemy_tag = self.store.session.query(AlchemyTag).get(tag.pk)
        if not alchemy_tag:
            message = _("No Tag with PK ‘{}’ was found.").format(tag.pk)
            self.store.logger.error(message)
            raise KeyError(message)
        alchemy_tag.name = tag.name

        try:
            self.store.session.commit()
        except IntegrityError as err:
            message = _(
                "An error occured! Are you sure that tag.name is not "
                "already present in the database? Error: '{}'."
            ).format(str(err))
            self.store.logger.error(message)
            raise ValueError(message)

        return alchemy_tag.as_hamster(self.store)

    # ***

    def remove(self, tag):
        """
        Delete a given tag.

        Args:
            tag (nark.Tag): Tag to be removed.

        Returns:
            None: If everything went alright.

        Raises:
            KeyError: If the ``Tag`` can not be found by the backend.
            ValueError: If tag passed does not have an pk.
        """

        self.store.logger.debug("Received: {!r}".format(tag))

        if not tag.pk:
            message = _(
                "Tag ‘{!r}’ has no PK. Are you trying to remove a new Tag?"
            ).format(tag)
            self.store.logger.error(message)
            raise ValueError(message)

        alchemy_tag = self.store.session.query(AlchemyTag).get(tag.pk)
        if not alchemy_tag:
            message = _("No Tag with PK ‘{}’ was found.").format(tag.pk)
            self.store.logger.error(message)
            raise KeyError(message)

        self.store.session.delete(alchemy_tag)
        self.store.session.commit()
        self.store.logger.debug("Deleted: {!r}".format(tag))

    # ***

    def get(self, pk, deleted=None):
        """
        Return a tag based on their pk.

        Args:
            pk (int): PK of the tag to be retrieved.

        Returns:
            nark.Tag: Tag matching given PK.

        Raises:
            KeyError: If no such PK was found.

        Note:
            We need this for now, as the service just provides pks, not names.
        """
        self.store.logger.debug("Received PK: ‘{}’".format(pk))

        if deleted is None:
            result = self.store.session.query(AlchemyTag).get(pk)
        else:
            query = self.store.session.query(AlchemyTag)
            query = query.filter(AlchemyTag.pk == pk)
            query = query_apply_true_or_not(query, AlchemyTag.deleted, deleted)
            results = query.all()
            assert len(results) <= 1
            result = results[0] if results else None

        if not result:
            message = _("No Tag with PK ‘{}’ was found.").format(pk)
            self.store.logger.error(message)
            raise KeyError(message)
        self.store.logger.debug("Returning: {!r}".format(result))
        return result.as_hamster(self.store)

    # ***

    def get_by_name(self, name, raw=False):
        """
        Return a tag based on its name.

        Args:
            name (str): Unique name of the tag.
            raw (bool): Wether to return the AlchemyTag instead.

        Returns:
            nark.Tag: Tag of given name.

        Raises:
            KeyError: If no tag matching the name was found.

        """
        self.store.logger.debug("Received name: ‘{}’ / raw: {}.".format(name, raw))

        try:
            result = self.store.session.query(AlchemyTag).filter_by(name=name).one()
        except NoResultFound:
            message = _("No Tag named ‘{}’ was found.").format(name)
            self.store.logger.debug(message)
            raise KeyError(message)

        if not raw:
            result = result.as_hamster(self.store)
            self.store.logger.debug("Returning: {!r}".format(result))
        return result

    # ***
    # *** gather() call-outs (used by get_all/get_all_by_usage).
    # ***

    @property
    def _gather_query_alchemy_cls(self):
        return AlchemyTag

    @property
    def _gather_query_order_by_name_col(self):
        return "tag"

    def _gather_query_requires_fact(self, qt, compute_usage):
        requires_fact_table = super(TagManager, self)._gather_query_requires_fact(
            qt, compute_usage
        )

        # (lb): I'm not sure the utility of querying Tag and sorting by
        # Activity, but we can easily support it.
        requires_fact_table = (
            requires_fact_table
            or qt.sort_cols_has_any("activity")
            or qt.sort_cols_has_any("category")
        )

        return requires_fact_table

    def _gather_query_start_aggregate(self, qt, agg_cols):
        # NOTE: We do not need a subquery here, unlike in gather_fact.py.
        #       - In _get_all_prepare_tags_subquery, which adds tags to the Fact
        #         gather query, the join-on-tags is performed in a subquery
        #         so that the (outer) group by aggregate functions do not
        #         inadvertently count each Fact more than once. In the case of
        #         a Fact gather, the tags names are being joined together for
        #         each Fact. But here, we actually want all the rows to be
        #         included in the aggregates, because we're counting usage the
        #         other way around -- how many Facts use each Tag; rather than
        #         how it is for the Fact gather, which is how many Tags of each
        #         name there are for each Fact. In any case, mostly just a note
        #         to self, because I looked at this and thought it smelled funny,
        #         because there's a subquery in the Fact gather query; but we
        #         don't need one here.
        query = self.store.session.query(AlchemyTag, *agg_cols)
        query = query.join(
            fact_tags,
            AlchemyTag.pk == fact_tags.columns.tag_id,
        )
        query = query.join(AlchemyFact)
        return query

    def query_criteria_filter_by_activities(self, query, qt):
        query, criteria = super(
            TagManager,
            self,
        ).query_criteria_filter_by_activities(query, qt)
        # (lb): I'm not sure the utility of querying Tag and sorting by
        # Activity, but we can easily support it.
        join_activity = bool(criteria) or qt.sort_cols_has_any("activity")
        if join_activity:
            query = query.join(AlchemyActivity)
        # (lb): I'm not sure if there's a way to check the query to see
        #       if it contains a table already; or whether the query is
        #       smart enough to discard multiple joins of the same table.
        #       So make a note to self that we joined the Activity table.
        qt.joined_activity = join_activity
        return query, criteria

    def query_criteria_filter_by_categories(self, query, qt):
        query, criteria = super(
            TagManager,
            self,
        ).query_criteria_filter_by_categories(query, qt)
        # (lb): I'm not sure the utility of querying Tag and sorting by
        # Category, but we can easily support it.
        join_category = bool(criteria) or qt.sort_cols_has_any("category")
        if join_category:
            if not qt.joined_activity:
                query = query.join(AlchemyActivity)
            query = query.join(AlchemyCategory)
        return query, criteria

    # ***
